//
// HyperoidCanvas - fullscreen canvas
//

import java.util.Random;
import java.util.Vector;
import java.io.InputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import javax.microedition.io.Connector;
import javax.microedition.io.file.FileConnection;
import javax.microedition.rms.RecordStore;
import javax.microedition.media.Manager;
import javax.microedition.media.Player;
import javax.microedition.media.control.ToneControl;
import javax.microedition.media.MediaException;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.game.GameCanvas;

/**
 * HyperoidCanvas contains the main game logic for Hyperoids.
 *
 * @see javax.microedition.lcdui.game.GameCanvas
 *
 * @author Edward Hutchins
 * Version: 2.0M  Copyright (C) 2006 Hutchins Software
 * portions Copyright (C) 2000 Russell Marks
 * portions Copyright (C) 1996 Rasmus Ekman
 * Version: 1.1  Copyright (C) 1990,91 Hutchins Software
 * This software is licenced under the GNU General Public Licence
 * Please read the associated legal documentation
 */

public class HyperoidCanvas
    extends GameCanvas
    implements Runnable
{
    private static final String Version = "2.0m";

    /**
     * Construct the canvas in fullscreen mode.
     * @param dpy Display passed from parent midlet.
     */

    public HyperoidCanvas( Display dpy )
    {
        // We don't want to capture keys normally
        super( true );
	rand = new Random();
        this.dpy = dpy;
        // We want a fullscreen canvas
        setFullScreenMode( true );
    }

    /**
     * Second-stage initialization, allows the majority of
     * initialization code to be potentially logged to a file.
     *
     * @param fps desired frames per second, or 0 for fastest possible rendering loop.
     */

    public void init( int fps )
    {
	if (logging)
	    openLogfile();

	loadPreferences( false );

	conPrint( "Hyperoids! version " + Version );
        // setup per-frame max sleep value
        fpsMillis = (fps > 0) ? (1000 / fps) : 0;
        conPrint( "fpsMillis = " + fpsMillis );

        clearScreen = true;
        displayedLevel = -1;
        displayedScore = -1;
        displayedLives = -1;
        displayedSheilds = -1;
        displayedBombs = -1;
    }

    // debug settings
    private static final boolean draw_keys = false;
    private static final boolean draw_fps = false;
    private static final boolean draw_cont_types = false;
    private static final boolean console = false;
    private static final boolean logging = false;
    private static final boolean run_as_thread = true;

    // Thread-control
    private static boolean running = false;
    private static boolean done = true;
    private static boolean enable_snd = true;
    private static long frameStartMillis;
    private static long elapsedMillis;
    private static long fpsMillis;

    // source of pseudo-randomness
    private static Random rand = null;

    // persistent storage
    private static AppPreferences prefs = null;

    // screen repainting controlls
    private static boolean clearScreen = true;
    private static int displayedLevel = -1;
    private static int displayedScore = -1;
    private static int displayedLives = -1;
    private static int displayedSheilds = -1;
    private static int displayedBombs = -1;

    private static Display dpy = null;

    // original game was on an 8Kx8K playfield, account for that with this shift value
    private static final int GAME_SCALE_BITS = 5;
    private static final int GAME_SCALE_ROUND = 1 << (GAME_SCALE_BITS - 1);
    private static int clipX = 0;
    private static int clipY = 0;

    // Key constants
    private static final int K_FIRE   = 0;
    private static final int K_UP     = 1;
    private static final int K_DOWN   = 2;
    private static final int K_LEFT   = 3;
    private static final int K_RIGHT  = 4;
    private static final int K_SHEILD = 5;
    private static final int K_BOMB   = 6;
    private static final int NUM_KEYS = 7;
    
    // Key array
    private int last_keystate = 0;
    private final boolean[] key = new boolean[NUM_KEYS];
    
    // for fixing the stupid backlight
    private long lastBacklightMillis = 0;

    // array of images for scores, icons etc.
    private static Image[] image = null;
    private static final int IMG_0 = 0;
    private static final int IMG_PLUS = IMG_0 + 10;
    private static final int IMG_LIFE = IMG_PLUS + 1;
    private static final int IMG_SHEILD = IMG_LIFE + 1;
    private static final int IMG_BOMB = IMG_SHEILD + 1;
    private static final int IMG_LEVEL = IMG_BOMB + 1;
    private static final int IMG_SCORE = IMG_LEVEL + 1;
    private static final int IMG_SIZE = IMG_SCORE + 1;
    private static final int IMGSIZE_BITS = 4;
    private static final int IMGSIZE_SIZE = (1 << IMGSIZE_BITS);
    private static final String[] imgnames =
    {
	"/Num0.png",
	"/Num1.png",
	"/Num2.png",
	"/Num3.png",
	"/Num4.png",
	"/Num5.png",
	"/Num6.png",
	"/Num7.png",
	"/Num8.png",
	"/Num9.png",
	"/Plus.png",
	"/Life.png",
	"/Shield.png",
	"/Bomb.png",
	"/Level.png",
	"/Score.png",
	"/Size.png"
    };

    // array of sound players for sound effects
    private static Player[] sound = null;
    private static final int SND_BONUSGOT = 0;
    private static final int SND_BSHOT = 1;
    private static final int SND_BWOUND = 2;
    private static final int SND_EXPLODE = 3;
    private static final int SND_EXPLODE2 = 4;
    private static final int SND_NEWBAD = 5;
    private static final int SND_NEWBONUS = 6;
    private static final int SND_PHIT = 7;
    private static final int SND_PSHOT = 8;
    private static final int SND_SWARMSPLIT = 9;
    private static final int SND_THRUST = 10;
    private static final int SND_TITLE = 11;
    private static final int SND_SHEILD_ON = 12;
    private static final int SND_SHEILD_OFF = 13;
    private static final String[] sndnames =
    {
	"/bonusgot.wav",
	"/bshot.wav",
	"/bwound.wav",
	"/explode.wav",
	"/explode2.wav",
	"/newbad.wav",
	"/newbonus.wav",
	"/phit.wav",
	"/pshot.wav",
	"/swarmsplit.wav",
	"/thrust.wav",
	"/title.wav",
	"/sheild_on.wav",
	"/sheild_off.wav"
    };

    // 256-entry cos/sin tables scaled to 1<<16 (fixed-point)
    private static final int DEG_BITS = 8;
    private static final int DEG_SIZE = 1 << DEG_BITS;
    private static final int DEG_MASK = DEG_SIZE - 1;
    // HACK HACK HACK precooked these to avoid using sin/cos (must match DEG_SIZE!)
    private static final int[] fxCos =
    {
        65536, 65516, 65457, 65358, 65220, 65043, 64826, 64571, 64276, 63943, 63571, 63162, 62714, 62228, 61705, 61144,
        60547, 59913, 59243, 58538, 57797, 57022, 56212, 55368, 54491, 53581, 52639, 51665, 50660, 49624, 48558, 47464,
        46340, 45189, 44011, 42806, 41575, 40319, 39039, 37736, 36409, 35061, 33692, 32302, 30893, 29465, 28020, 26557,
        25079, 23586, 22078, 20557, 19024, 17479, 15923, 14359, 12785, 11204, 9616, 8022, 6423, 4821, 3215, 1608,
        0, -1608, -3215, -4821, -6423, -8022, -9616, -11204, -12785, -14359, -15923, -17479, -19024, -20557, -22078, -23586,
        -25079, -26557, -28020, -29465, -30893, -32302, -33692, -35061, -36409, -37736, -39039, -40319, -41575, -42806, -44011, -45189,
        -46340, -47464, -48558, -49624, -50660, -51665, -52639, -53581, -54491, -55368, -56212, -57022, -57797, -58538, -59243, -59913,
        -60547, -61144, -61705, -62228, -62714, -63162, -63571, -63943, -64276, -64571, -64826, -65043, -65220, -65358, -65457, -65516,
        -65536, -65516, -65457, -65358, -65220, -65043, -64826, -64571, -64276, -63943, -63571, -63162, -62714, -62228, -61705, -61144,
        -60547, -59913, -59243, -58538, -57797, -57022, -56212, -55368, -54491, -53581, -52639, -51665, -50660, -49624, -48558, -47464,
        -46340, -45189, -44011, -42806, -41575, -40319, -39039, -37736, -36409, -35061, -33692, -32302, -30893, -29465, -28020, -26557,
        -25079, -23586, -22078, -20557, -19024, -17479, -15923, -14359, -12785, -11204, -9616, -8022, -6423, -4821, -3215, -1608,
        0, 1608, 3215, 4821, 6423, 8022, 9616, 11204, 12785, 14359, 15923, 17479, 19024, 20557, 22078, 23586,
        25079, 26557, 28020, 29465, 30893, 32302, 33692, 35061, 36409, 37736, 39039, 40319, 41575, 42806, 44011, 45189,
        46340, 47464, 48558, 49624, 50660, 51665, 52639, 53581, 54491, 55368, 56212, 57022, 57797, 58538, 59243, 59913,
        60547, 61144, 61705, 62228, 62714, 63162, 63571, 63943, 64276, 64571, 64826, 65043, 65220, 65358, 65457, 65516,
    };
    private static final int[] fxSin =
    {
        0, 1608, 3215, 4821, 6423, 8022, 9616, 11204, 12785, 14359, 15923, 17479, 19024, 20557, 22078, 23586,
        25079, 26557, 28020, 29465, 30893, 32302, 33692, 35061, 36409, 37736, 39039, 40319, 41575, 42806, 44011, 45189,
        46340, 47464, 48558, 49624, 50660, 51665, 52639, 53581, 54491, 55368, 56212, 57022, 57797, 58538, 59243, 59913,
        60547, 61144, 61705, 62228, 62714, 63162, 63571, 63943, 64276, 64571, 64826, 65043, 65220, 65358, 65457, 65516,
        65536, 65516, 65457, 65358, 65220, 65043, 64826, 64571, 64276, 63943, 63571, 63162, 62714, 62228, 61705, 61144,
        60547, 59913, 59243, 58538, 57797, 57022, 56212, 55368, 54491, 53581, 52639, 51665, 50660, 49624, 48558, 47464,
        46340, 45189, 44011, 42806, 41575, 40319, 39039, 37736, 36409, 35061, 33692, 32302, 30893, 29465, 28020, 26557,
        25079, 23586, 22078, 20557, 19024, 17479, 15923, 14359, 12785, 11204, 9616, 8022, 6423, 4821, 3215, 1608,
        0, -1608, -3215, -4821, -6423, -8022, -9616, -11204, -12785, -14359, -15923, -17479, -19024, -20557, -22078, -23586,
        -25079, -26557, -28020, -29465, -30893, -32302, -33692, -35061, -36409, -37736, -39039, -40319, -41575, -42806, -44011, -45189,
        -46340, -47464, -48558, -49624, -50660, -51665, -52639, -53581, -54491, -55368, -56212, -57022, -57797, -58538, -59243, -59913,
        -60547, -61144, -61705, -62228, -62714, -63162, -63571, -63943, -64276, -64571, -64826, -65043, -65220, -65358, -65457, -65516,
        -65536, -65516, -65457, -65358, -65220, -65043, -64826, -64571, -64276, -63943, -63571, -63162, -62714, -62228, -61705, -61144,
        -60547, -59913, -59243, -58538, -57797, -57022, -56212, -55368, -54491, -53581, -52639, -51665, -50660, -49624, -48558, -47464,
        -46340, -45189, -44011, -42806, -41575, -40319, -39039, -37736, -36409, -35061, -33692, -32302, -30893, -29465, -28020, -26557,
        -25079, -23586, -22078, -20557, -19024, -17479, -15923, -14359, -12785, -11204, -9616, -8022, -6423, -4821, -3215, -1608,
    };

    /**
     * A simple hit-testing rectangle for checking rectangle-rectangle
     * and rectangle-point collisions.
     */

    private class HitRect
    {
	int		top, left, bottom, right;

	/**
	 * reset the rect to empty
	 */
	
	public final void empty()
	{
	    top = left = 0x7fffffff;
	    bottom = right = -0x7fffffff;
	}
	
	/**
	 * create a new empty rect
	 */
	
	public HitRect()
	{
	    empty();
	}
	
	/**
	 * sets the rectangle to be centered around Pt extending +/- nSize units.
	 * @param Pt int[2], x in [0], y in [1].
	 * @param nSize size of rectangle to create.
	 */
	
	public final void set( int[] Pt, int nSize )
	{
	    top = Pt[1] - nSize;
	    bottom = Pt[1] + nSize;
	    left = Pt[0] - nSize;
	    right = Pt[0] + nSize;
	}
	
	/**
	 * creates a rectangle centered around Pt extending +/- nSize units.
	 * @param Pt int[2], x in [0], y in [1].
	 * @param nSize size of rectangle to create.
	 */
	
	public HitRect( int[] Pt, int nSize )
	{
	    set( Pt, nSize );
	}
	
	/**
	 * extends the rectangle to encompass Pt
	 * @param Pt int[2], x in [0], y in [1].
	 */
	
	public final void grow( int[] Pt )
	{
	    if (Pt[0] < left) left = Pt[0];
	    if (Pt[0] > right) right = Pt[0];
	    if (Pt[1] < top) top = Pt[1];
	    if (Pt[1] > bottom) bottom = Pt[1];
	}
	
	/**
	 * tests Pt against the rectangle
	 * @param Pt int[2], x in [0], y in [1].
	 * @return true if the point is in the rectangle
	 */
	
	public final boolean hits( int[] Pt )
	{
	    return( (Pt[0] >= left) && (Pt[0] <= right) && (Pt[1] >= top) && (Pt[1] <= bottom) );
	}
	
	/**
	 * tests another rectangle r against the rectangle
	 * @param r other rectangle to test
	 * @return true if the rectangles intersect
	 */
	
	public final boolean hits( HitRect r )
	{
	    return( (right >= r.left) && (left <= r.right) && (bottom >= r.top) && (top <= r.bottom) );
	}
    }

    // must match array decls below
    private static final int MAX_PTS = 16;

    /**
     * main hyperoid object class (pretty much everything drawn on the screen uses this object)
     * since we want to avoid using the allocator during play, all objects are just mashed together
     * into this one class (which is fairly unchanged from the original C Windows 3.1 implementaion)
     */

    private class HyperObj
    {
        int[]           Pos = { 0, 0 };     // position of center of object
        int[]           Vel = { 0, 0 };     // velocity in logical units/update
        int             nMass;              // mass of object
        int             nDir;               // direction in degrees
        int             nSpin;              // angular momentum degrees/update
        int             nCount;             // used by different objects (0 => not drawn)
        long		lDelay;             // used by different objects as an elapsed time interval
        int             nColor;             // 0xRRGGBB color
        int             nPts;               // number of points in object

	// points making up an object (radial coordinates)
        int[][]         Pts =
        {
            { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, 
            { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, 
            { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, 
            { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }
        };

	// last plotted location (cartesian coordinates)
	int[][]         Old =
        {
            { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, 
            { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, 
            { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, 
            { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }
        };
    }

    // these templates are actually angle [0] and length [0] in a radial coordinate system
    private static final int[][] PlayerShape =
    {
        {0, 0}, {160, 150}, {0, 250}, {96, 150}, {0, 0}
    };
    private static final int[][] SpinnerShape =
    {
        {160, 150}, {224, 100}, {96, 100}, {32, 150}, {160, 150}
    };
    private static final int[][] SwarmerShape =
    {
        {0, 100}, {64, 100}, {128, 100}, {192, 100}, {0, 100}
    };
    private static final int[][] HunterShape =
    {
        {160, 150}, {0, 250}, {192, 30}, {64, 30},
        {0, 250}, {96, 150}, {128, 150}, {160, 150}
    };
    private static final int[][] BonusShape =
    {
        {0, 150}, {102, 150}, {205, 150}, {51, 150}, {154, 150}, {0, 150}
    };
    // these parts map to "abcdefghijklm"
    // Ie, map to points in a LED-style grid:
    //   a b c
    //   d n e
    //   f g h
    //   i o j
    //   k l m
    private static final int[][] LetterPart =
    {
        {83, 572},   {64, 512}, {45, 572},
	{96, 362},              {32, 362},
        {128, 256},   {0, 0},   {0, 256},
        {160, 362},             {224, 362},
	{173, 572}, {192, 512}, {211, 572},
	// added these later
	{64, 256}, {192, 256}
    };
    // here's the vector font
    private static final String[] NumberDesc =
    {
        "cakmck",       // 0
        "dbl",          // 1
        "abekm",        // 2
        "abegjlk",      // 3
        "mcfh",         // 4
        "cbfgjlk",      // 5
        "bdiljgi",      // 6
        "acgl",         // 7
        "bdjlieb",      // 8
        "ljebdge"       // 9
    };
    private static final String[] LetterDesc =
    {
        "kdbemhf",      // A
        "kabegjlk",     // B
        "cbflm",        // C
        "kabejlk",      // D
        "cafgfkm",      // E
        "cafgfk",       // F
        "bdiljhg",      // G
        "kafhcm",       // H
        "bl",           // I
        "cjli",         // J
        "akcgm",        // K
        "akm",          // L
        "kagcm",        // M
        "kamc",         // N
        "bdiljeb",      // O
        "kabegf",       // P
        "mlidbejl",     // Q
        "kabegfgm",     // R
        "ebdjli",       // S
        "lbac",         // T
        "ailjc",        // U
        "alc",          // V
        "akgmc",        // W
        "amgkc",        // X
        "aglgc",        // Y
        "ackm"          // Z
    };
    //   a b c
    //   d n e
    //   f g h
    //   i o j
    //   k l m
    private static final String[] PuncDesc =
    {
	"-ll",		// .
	"ok",		// ,
	"fh",		// -
	"fhgno",	// +
	"ck",		// /
	"cnom",		// (
	"anok",		// )
	"cblm",		// [
	"ablk",		// [
	"-no",		// :
	"-nok"		// ;
    };
    private static final String puncChars = ".,-+/()[]:;";

    /**
     * Class describing a menu item, full menus are just Vectors of these HyperMenuItems
     */

    private class HyperMenuItem
    {
	public int	    nSize;	// letter/glyph size
	public int	    nColor;	// 0xrrggbb
	public int	    nCmd;	// what MENU_* command this corresponds to (0 == none)
	public String	    text;	// text corresponding to Vector below
	public Vector	    objs;	// objects to draw

	/**
	 * build a menu item from a string description 
	 *
	 * @param desc string describing menu item, with optional !cmd; @size; and #color descriptors
	 * example menu descriptor: "!45;@250;#ff0000;Hello" => cmd 45, size 250, red text "Hello"
	 */
	public HyperMenuItem( String desc )
	{
	    nSize = MENU_DEFAULT_SIZE;
	    nColor = MENU_DEFAULT_COLOR;
	    nCmd = 0;
	    text = "";
	    objs = new Vector();

	    int i = 0;
	    while (i < desc.length())
	    {
		switch (desc.charAt( i ))
		{
		case '!': // sets nCmd
		case '@': // sets nSize
		case '#': // sets nColor from hex
		    {
			char c = desc.charAt( i );
			int vbegin = ++i;
			String s = null;
			while (i < desc.length())
			{
			    if (desc.charAt( i ) == ';')
			    {
				s = desc.substring( vbegin, i++ );
				break;
			    }
			    ++i;
			}
			if (s == null) throw new Error( "Badly formed menu descriptor " + desc );
			switch (c)
			{
			case '!': nCmd = Integer.parseInt( s ); break;
			case '@': nSize = Integer.parseInt( s ); break;
			case '#': nColor = Integer.parseInt( s, 16 ); break;
			}
		    }
		    break;
		default:
		    text = desc.substring( i ).trim();
		    i = desc.length();
		    break;
		}
	    }
	}
    }

    //
    // game engine constants
    //

    // sizes
    private static final int SAFE_ZONE_SIZE = 100;
    private static final int LETTER_SIZE = 1024;
    private static final int LETTER_SCALE = 256;
    private static final int MENU_DEFAULT_SIZE = 128;
    private static final int MENU_SPACING = 1;
    private static final int PLAYER_HITSIZE = 200;
    private static final int BONUS_HITSIZE = 150;
    private static final int SPINNER_HITSIZE = 150;
    private static final int HUNTER_HITSIZE = PLAYER_HITSIZE;
    private static final int SWARMER_HITSIZE = 150;
    private static final int SWARMER_DELTASIZE = 10;

    // speeds
    private static final int PLAYER_SPEEDLIMIT = 8192;

    // colors
    private static final int MENU_DEFAULT_COLOR = 0x20ff20;
    private static final int PLAYER_COLOR = 0xffffff;
    private static final int PLAYER_COLOR_2 = 0xc0c0c0;
    private static final int PLAYER_COLOR_1 = 0x7f7f7f;
    private static final int PLAYER_SHEILDCOLOR = 0x7fff7f;
    private static final int SHOT_COLOR = 0x00ffff;    
    private static final int FLAME_YELLOW = 0xffff00;
    private static final int FLAME_ORANGE = 0xff7f00;
    private static final int FLAME_RED = 0xff0000;
    private static final int FLAME_DKRED = 0x7f0000;
    private static final int TEXT_YELLOW = 0xc0a000;
    private static final int TEXT_RED = 0xff0000;
    private static final int TEXT_BLUE = 0x1020ff;
    private static final int[] BONUS_COLORS =
    {
	0xffffff, // count == 0, not used
	0xff0000, // points
	0x00ff00, // sheilds
	0x0000ff, // bombs
	0xffff00  // extra life
    };
    private static final int PLAYER_MESSAGE_COLOR = 0x00ff30;
    private static final int HUNTER_COLOR = 0x00ffff;
    private static final int HUNTER_COLOR_1 = 0x007f7f;
    private static final int HUNTER_SHOT_COLOR = 0x00ff00;
    private static final int SPINNER_COLOR = 0xff00ff;
    private static final int SWARMER_COLOR = 0x208010;

    // how much things weigh
    private static final int PLAYER_MASS = 256;
    private static final int SHOT_MASS = 8;
    private static final int BONUS_MASS = 1;
    private static final int HUNTER_MASS = PLAYER_MASS;
    private static final int SWARMER_MASS = 32;
    private static final int SPINNER_MASS = 50;
    private static final int SPINNER_MASS_DELTA = 25;

    // durations
    private static final int MAX_TONE_QUEUE = 30;
    private static final int PLAYER_SHOTDELAY = 5;
    private static final int RESTART_GAME_DELAY = 500;
    private static final int RESTART_ATTRACT_DELAY = 5000;
    private static final int RESTART_LEVEL_DELAY = 3000;
    private static final int RESTART_NEXTLEVEL_DELAY = 3000;
    private static final int SHEILD_TIME = 1000;
    private static final int BONUS_INTERVAL = 33000;
    private static final int BONUS_CHANGE_MINDELAY = 2000;
    private static final int BONUS_CHANGE_RANDDELAY = 3000;
    private static final int HUNTER_INTERVAL = 25000;
    private static final int HUNTER_RANDINTERVAL = 3000;
    private static final int HUNTER_LEVEL_QUICKENER = 500;
    private static final int HUNTER_MININTERVAL = 5000;
    private static final int HUNTER_CHANGE_MINDELAY = 66;
    private static final int HUNTER_CHANGE_RANDDELAY = 100;
    private static final int SWARMER_INTERVAL = 20000;
    private static final int SWARMER_RANDINTERVAL = 8000;
    private static final int SWARMER_LEVEL_QUICKENER = 800;
    private static final int SWARMER_MININTERVAL = 5000;
    private static final int SWARMER_CHANGE_MINDELAY = 4000;
    private static final int SWARMER_CHANGE_RANDDELAY = 8000;
    private static final int SPINNER_INTERVAL = 10000;
    private static final int SPINNER_RANDINTERVAL = 10000;
    private static final int SPINNER_LEVEL_QUICKENER = 1000;
    private static final int SPINNER_MININTERVAL = 5000;
    private static final int ATTRACT_MENU_INTERVAL = 5000;
    private static final int MENU_DELAY = 250;
    private static final int LICENSE_DELAY = 15;

    // application modes
    private static final int APP_ATTRACT = 0;
    private static final int APP_PLAY = 1;
    private static final int APP_MENU = 2;
    
    // restart modes
    private static final int RESTART_GAME = 0;
    private static final int RESTART_ATTRACT = 1;
    private static final int RESTART_LEVEL = 2;
    private static final int RESTART_NEXTLEVEL = 3;

    // other scalar limits
    private static final int EXTRA_LIFE_SCORE = 100000;
    private static final int MAX_HUNTERS = 5;
    private static final int MAX_SWARMERS = 50;
    private static final int MAX_SPINNERS = 5;
    private static final int MAX_HIGHSCORES = 5;

    // menu commands
    private static final int MENU_SET_HIGHSCORE = 1;
    private static final int MENU_BACK = 2;
    private static final int MENU_ABOUT = 3;
    private static final int MENU_HELP = 4;
    private static final int MENU_OPTIONS = 5;
    private static final int MENU_EXIT = 6;
    private static final int MENU_RESUME = 7;
    private static final int MENU_NEWGAME = 8;
    private static final int MENU_LICENSE = 9;
    private static final int MENU_OPT_SOUND = 10;
    private static final int MENU_OPT_RESET = 11;

    // attract menu types
    private static final int ATTRACT_MENU_TITLE = 0;
    private static final int ATTRACT_MENU_HELP= 1;
    private static final int ATTRACT_MENU_HIGHSCORE = 2;
    private static final int NUM_ATTRACT_MENUS = 3;

    // the license
    private static final String LICENSE[] =
    {
	"This program is free software;",
	"you can redistribute it and/or",
	"modify it under the terms of the",
	"GNU General Public License as",
	"published by the Free Software",
	"Foundation; either version 1, or",
	"(at your option) any later version.",
	"",
	"This program is distributed in",
	"the hope that it will be useful,",
	"but WITHOUT ANY WARRANTY; without",
	"even the implied warranty of",
	"MERCHANTABILITY or FITNESS FOR A",
	"PARTICULAR PURPOSE. See the GNU",
	"General Public License for more",
	"details:",
	"",
	"www.gnu.org/copyleft/gpl.html",
	"",
	"or write to:",
	"Free Software Foundation",
	"51 Franklin Street, Fifth Floor",
	"Boston, MA 02110-1301",
	"USA"
    };

    // special highscore menu (3rd item dynamically generated)
    private static final String[] HighScoreMenuDesc =
    {
	"@250;#ff7f00;*NEW HIGH SCORE*",
	"",
	null
    };
    private static final int HIGHSCORE_MENU_ENTRY = 2;

    // top-level menu for APP_MENU mode
    private static final String[] RootMenuDesc =
    {
	"@300;#ffff00;*HYPEROIDS*",
	"",
	"!" + String.valueOf( MENU_RESUME ) + ";resume",
	"!" + String.valueOf( MENU_NEWGAME ) + ";new game",
	"!" + String.valueOf( MENU_HELP ) + ";help",
	"!" + String.valueOf( MENU_OPTIONS ) + ";options",
	"!" + String.valueOf( MENU_ABOUT ) + ";credits",
	"!" + String.valueOf( MENU_EXIT ) + ";exit"
    };

    private static final String[] OptionMenuDesc =
    {
	"@300;#ffff00;OPTIONS",
	"",
	"!" + String.valueOf( MENU_OPT_SOUND ) + ";sound",
	"!" + String.valueOf( MENU_OPT_RESET ) + ";reset",
	"",
	"!" + String.valueOf( MENU_BACK ) + ";back"
    };
    private static final int OPT_SOUND_MENU_ENTRY = 2;

    private static final String[] AboutMenuDesc =
    {
	"@300;#ffff00;*HYPEROIDS*",
	"",
	"#808080;version " + Version,
	"#808080;(C) 1990-2006",
	"",
	"@150;#c0c0c0;Edward Hutchins",
	"@150;#c0c0c0;Rasmus Ekman",
	"@150;#c0c0c0;Russell Marks",
	"",
	"!" + String.valueOf( MENU_BACK ) + ";back",
	"!" + String.valueOf( MENU_LICENSE ) + ";license"
    };

    private static final String[] HelpMenuDesc =
    {
	"@200;HELP",
	"",
	"up ..... rev thrust",
	"down ....... thrust",
	"left/right ... turn",
	"game a .... sheilds",
	"game b ....... bomb",
	"",
	"extra life: 100,000",
	"",
	"play continues from last",
	"level (use the menu to",
	"start a new game)",
	"",
	"!" + String.valueOf( MENU_BACK ) + ";back",
    };

    // attract "menus"
    private static final String[] AttractMenuTitleDesc =
    {
	"@300;#ffff00;*HYPEROIDS*",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"press fire",
	"to start",
    };
    private static final String[] AttractMenuHelpDesc =
    {
	"#c0c0c0;- enemies -",
	"#" + Integer.toHexString( HUNTER_COLOR ) +    ";^ ..... hunter",
	"#" + Integer.toHexString( SPINNER_COLOR ) +   ";$ .... spinner",
	"#" + Integer.toHexString( SWARMER_COLOR ) +   ";% .... swarmer",
	"",
	"#c0c0c0;- bonuses -",
	"#" + Integer.toHexString( BONUS_COLORS[4] ) + ";* ....... life",
	"#" + Integer.toHexString( BONUS_COLORS[3] ) + ";* ....... bomb",
	"#" + Integer.toHexString( BONUS_COLORS[2] ) + ";* ..... sheild",
	"#" + Integer.toHexString( BONUS_COLORS[1] ) + ";* ..... points"
    };
    private static final String[] AttractMenuHighScoreDesc =
    {
	"@200;#ffff00;hall of fame",
	"",
	// these are filled in when loading prefs and setting highscores
	"@160;EAH - 5000",
	"@160;EAH - 4000",
	"@160;EAH - 3000",
	"@160;EAH - 2000",
	"@160;EAH - 1000",
    };
    private static final int FIRST_HIGHSCORE_MENU_ENTRY = 2;
    private static final String[][] AttractMenuDesc =
    {
	AttractMenuTitleDesc,
	AttractMenuHelpDesc,
	AttractMenuHighScoreDesc
    };

    /**
     * simple high-score record
     */

    private class HighScore
    {
	public String	name;
	public int	score;

	/**
	 * set the HighScore record's name and score
	 * 
	 * @param name player's name
	 * @param score player's score
	 */

	public HighScore( String name, int score ) { this.name = name; this.score = score; }
	
	/**
	 * format the high score for storage in a user preference value
	 * @return formatted string
	 */

	public String getText() { return( name + "," + score ); }
	
	/**
	 * format the high score for use in a HyperMenuItem
	 * @return formatted string
	 */

	public String getMenuText() { return( "@160;" + name + " - " + score ); }
    }

    // preference keys
    private static final String highScoreChars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"; // legal high score name characters
    private static final String PREFKEY_LASTHSNAME = "lasthsname";
    private static final String PREFKEY_HS = "hs"; // + index
    private static final String PREFKEY_SOUND = "sound"; // enable_snd

    //
    // game engine globals
    //

    // game logic globals
    private static int gameMode = APP_ATTRACT; // what's going on
    private static int gameAttractMenu = 0; // current attract menu
    private static int gameLevel = 0; // what level the player is currently at
    private static int gameScore = 0; // total score for this game
    private static int gameLastLifeScore = 0; // score of last extra life bonus
    private static int gameLives = 3; // lives remaining
    private static int gameSheilds = 3; // sheilds remaining
    private static int gameBombs = 3; // smartbombs remaining
    private static int gameBadGuys = 0; // number of roids left to break
    private static boolean gameCreateBadGuys = false; // mode where new badguys can appear
    private static boolean gameSetHighScore = false; // set the highscore on next restart
    private static HighScore[] gameHighScore = new HighScore[MAX_HIGHSCORES]; // best score so far

    // player state
    private static int playerShotDelay = 0; // frames between shots
    private static int playerBombing = 0; // bombing in progress
    private static boolean playerBombDebounce = false; // bombing key held down
    private static char lastHighScoreName[] = { 'A', 'A', 'A' }; // last used highscore name

    // menu state
    private static int menuSelection = -1; // currently active menu item
    private static int menuHighScoreChar = 0; // currently active menu character in high score
    private static int menuResumeMode; // mode we exit APP_MENU back to
    private static boolean menuCreateRoot = false; // root menu is pending creation
    private static boolean menuFireKeyDebounce = false; // last state of fire key

    // timers
    private static long restartMillis; // if > 0, time of next game restart
    private static long safeMillis; // if > 0, time that sheilds expire
    private static long nextBonusMillis; // if > 0, time that next bonus is granted
    private static long nextHunterMillis; // if > 0, time next hunter gets created
    private static long nextSwarmerMillis; // if > 0, time next swarmer gets created
    private static long nextSpinnerMillis; // if > 0, time next spinner gets created
    private static long toneEndMillis; // if > 0, approx time queued tones will end
    private static long attractSwitchMillis; // if > 0, time to next attract menu switch
    private static long menuDelayMillis; // if > 0, time to next keyboard repeat
    private static long menuModeStartMillis; // if > 0, time we entered APP_MENU
    
    // for the license thingy
    private static int licenseLine = 0; // next line to print
    private static long licenseTicks = 0; // countdown to next license line printout

    // sets of things on the playfield
    private static HyperObj Player = null;
    private static final Vector FreeList = new Vector();
    private static final Vector RoidList = new Vector();
    private static final Vector ShotList = new Vector();
    private static final Vector FlameList = new Vector();
    private static final Vector SpinnerList = new Vector();
    private static final Vector HunterList = new Vector();
    private static final Vector HunterShotList = new Vector();
    private static final Vector SwarmerList = new Vector();
    private static final Vector LetterList = new Vector();
    private static final Vector BonusList = new Vector();
    private static final Vector MenuList = new Vector();

    // bounding rect for player shots to speed up hit-testing
    private final HitRect shotRect = new HitRect();
    // scratch rectangle used by most hit testing
    private final HitRect rect = new HitRect();

    /**
     * create a new object (checks the freelist first)
     * 
     * @param Shape optional shape to initialize object with
     * @return new HyperObj
     */

    private final HyperObj newObj( int[][] Shape )
    {
        HyperObj o = null;
        if (FreeList.size() > 0)
        {
            o = (HyperObj)FreeList.lastElement();
            FreeList.removeElementAt( FreeList.size() - 1 );
        }
        else o = new HyperObj();
        if (Shape != null)
        {
            o.nPts = Shape.length;
            for (int i = 0; i < o.nPts; ++i)
            {
                o.Pts[i][0] = Shape[i][0];
                o.Pts[i][1] = Shape[i][1];
            }
        }
        return( o );
    }

    /**
     * remove the object at idx and put it in the freelist
     * (note: iterating forward through v should repeat idx!)
     * 
     * @param v Vector to remove object from
     * @param idx position in Vector to be removed
     */

    private final void delObj( Vector v, int idx )
    {
        HyperObj o = (HyperObj)v.elementAt( idx );
        v.removeElementAt( idx ); // moves last to idx
        FreeList.addElement( o );
    }

    /**
     * just redraw an object (no update)
     * if nCount == 0, erases object
     * 
     * @param g Graphics context
     * @param o object to redraw
     */

    private final void drawObj( Graphics g, HyperObj o )
    {
	g.setColor( (o.nCount > 0) ? o.nColor : 0x000000 );
	if (o.nPts == 1)
	{
	    // single point
	    g.fillRect( o.Old[0][0], o.Old[0][1], 1, 1 );
	}
	else if (o.nPts > 1)
	{
	    // open line segment
            for (int i = 1; i < o.nPts; ++i)
                g.drawLine( o.Old[i - 1][0], o.Old[i - 1][1], o.Old[i][0], o.Old[i][1] );
	}
	else if (o.nPts < 0)
	{
	    // point group
            for (int i = 0; i < -o.nPts; ++i)
		g.fillRect( o.Old[i][0], o.Old[i][1], 1, 1 );
	}
    }

    /**
     * update and redraw an object
     * if nCount == 0, erases object
     * 
     * @param g Graphics context
     * @param o object to update
     */

    private final void updateObj( Graphics g, HyperObj o )
    {
	int x = (o.Pos[0] += o.Vel[0]);
	int y = (o.Pos[1] += o.Vel[1]);

        // keep objects in the classic torroidally-connected *steroids space
        if      (x < -clipX) o.Pos[0] = (x += clipX << 1);
	else if (x >  clipX) o.Pos[0] = (x -= clipX << 1);
        if      (y < -clipY) o.Pos[1] = (y += clipY << 1);
	else if (y >  clipY) o.Pos[1] = (y -= clipY << 1);

	if (o.nPts == 1)
	{
	    // single point
            g.setColor( 0x000000 );
            g.fillRect( o.Old[0][0], o.Old[0][1], 1, 1 );
            if (o.nCount > 0)
            {
                o.Old[0][0] = x >> GAME_SCALE_BITS;
                o.Old[0][1] = y >> GAME_SCALE_BITS;
                g.setColor( o.nColor );
                g.fillRect( o.Old[0][0], o.Old[0][1], 1, 1 );
            }
	}
	else if (o.nPts > 1)
	{
	    // open line segment
            g.setColor( 0x000000 );
            for (int i = 1; i < o.nPts; ++i)
                g.drawLine( o.Old[i - 1][0], o.Old[i - 1][1], o.Old[i][0], o.Old[i][1] );
            if (o.nCount > 0)
            {
                // update the object with its new position and rotation
                int nDir = (o.nDir += o.nSpin);
                for (int nCnt = o.nPts - 1; nCnt >= 0; --nCnt)
                {
                    int nDeg = (((o.Pts[nCnt][0] << GAME_SCALE_BITS) + nDir + GAME_SCALE_ROUND) >> GAME_SCALE_BITS) & DEG_MASK;
                    int nLen = o.Pts[nCnt][1];
                    // convert back to integer values (bake in a scale factor to account for PC->cellphone screen size)
                    o.Old[nCnt][0] = (((x + GAME_SCALE_ROUND) << 16) + (nLen * fxCos[nDeg])) >> (GAME_SCALE_BITS + 16);
                    o.Old[nCnt][1] = (((y + GAME_SCALE_ROUND) << 16) + (nLen * fxSin[nDeg])) >> (GAME_SCALE_BITS + 16);
                }
                g.setColor( o.nColor );
                for (int i = 1; i < o.nPts; ++i)
                    g.drawLine( o.Old[i - 1][0], o.Old[i - 1][1], o.Old[i][0], o.Old[i][1] );
            }
	}
	else if (o.nPts < 0)
	{
	    // point group
            g.setColor( 0x000000 );
            for (int i = 0; i < -o.nPts; ++i)
		g.fillRect( o.Old[i][0], o.Old[i][1], 1, 1 );
            if (o.nCount > 0)
            {
                // update the object with its new position and rotation
                int nDir = (o.nDir += o.nSpin);
                for (int nCnt = -o.nPts - 1; nCnt >= 0; --nCnt)
                {
                    int nDeg = (((o.Pts[nCnt][0] << GAME_SCALE_BITS) + nDir + GAME_SCALE_ROUND) >> GAME_SCALE_BITS) & DEG_MASK;
                    int nLen = o.Pts[nCnt][1];
                    // convert back to integer values (bake in a scale factor to account for PC->cellphone screen size)
                    o.Old[nCnt][0] = (((x + GAME_SCALE_ROUND) << 16) + (nLen * fxCos[nDeg])) >> (GAME_SCALE_BITS + 16);
                    o.Old[nCnt][1] = (((y + GAME_SCALE_ROUND) << 16) + (nLen * fxSin[nDeg])) >> (GAME_SCALE_BITS + 16);
                }
                g.setColor( o.nColor );
                for (int i = 0; i < -o.nPts; ++i)
		    g.fillRect( o.Old[i][0], o.Old[i][1], 1, 1 );
            }
	}
    }

    /**
     * accelerate an object along a direction
     * 
     * @param o object to accelerate
     * @param nDir direction to accelerate along
     * @param nAccel amount of acceleration
     */

    private final void accelObj( HyperObj o, int nDir, int nAccel )
    {
	nDir = (nDir + GAME_SCALE_ROUND) >> GAME_SCALE_BITS;
	o.Vel[0] += (nAccel * fxCos[nDir & DEG_MASK] + 0x8000) >> 16;
	o.Vel[1] += (nAccel * fxSin[nDir & DEG_MASK] + 0x8000) >> 16;
    }

    private static final int HIT_PARTICLES = 5;
    private static final int HIT_DIRCHANGE = (DEG_SIZE << GAME_SCALE_BITS) / HIT_PARTICLES;

    /**
     * add special effects for when something is hit
     * 
     * @param o object to "hit"
     */

    private final void hitObj( HyperObj o )
    {
	int nDir = o.nDir + rand.nextInt( HIT_DIRCHANGE );
	for (int i = 0; i < HIT_PARTICLES; ++i)
	{
	    HyperObj flame = newObj( null );
	    flame.Pos[0] = o.Pos[0];
	    flame.Pos[1] = o.Pos[1];
	    flame.Vel[0] = o.Vel[0];
	    flame.Vel[1] = o.Vel[1];
	    // distribute around a circle
	    flame.nDir = nDir;
	    nDir += HIT_DIRCHANGE;
	    flame.nSpin = 0;
	    flame.nCount = 10 + rand.nextInt( 8 );
	    flame.nColor = FLAME_YELLOW;
	    flame.nPts = 1;
	    flame.Pts[0][0] = flame.Pts[0][1] = 0;
	    accelObj( flame, flame.nDir, 50 - flame.nCount );
	    FlameList.addElement( flame );
	}
    }

    /**
     * blow an object up
     * 
     * @param g Graphics context
     * @param o object to explode
     */

    private final void explodeObj( Graphics g, HyperObj o )
    {
	int nSize = o.nPts;

	// erase the object (caller must set o.nCount <= 0)
	updateObj( g, o );

	for (int i = 0; i < nSize; ++i)
	{
	    if (rand.nextInt( 2 ) > 0) continue;
	    HyperObj flame = newObj( null );
	    flame.Pos[0] = o.Pos[0];
	    flame.Pos[1] = o.Pos[1];
	    flame.Vel[0] = o.Vel[0];
	    flame.Vel[1] = o.Vel[1];
	    flame.nDir = o.nDir + (i * DEG_SIZE / nSize) + rand.nextInt( 32 );
	    flame.nSpin = rand.nextInt( (30 << GAME_SCALE_BITS) + 1 ) - (15 << GAME_SCALE_BITS);
	    flame.nCount = 25 + rand.nextInt( 16 );
	    flame.nColor = o.nColor;
	    flame.nPts = 2;
	    flame.Pts[0][0] = o.Pts[i][0];
	    flame.Pts[0][1] = o.Pts[i][1];
	    if (i == (nSize - 1))
	    {
		flame.Pts[1][0] = o.Pts[0][0];
		flame.Pts[1][1] = o.Pts[0][1];
	    }
	    else 
	    {
		flame.Pts[1][0] = o.Pts[i + 1][0];
		flame.Pts[1][1] = o.Pts[i + 1][1];
	    }
	    accelObj( flame, flame.nDir, 60 - flame.nCount );
	    FlameList.addElement( flame );
	}
	hitObj( o );
    }

    /**
     * draw an entire list of objects
     * 
     * @param g Graphics context
     * @param list Vector of objects to call drawObj with
     */

    private final void drawObjects( Graphics g, Vector list )
    {
	for (int i = list.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)list.elementAt( i );
	    drawObj( g, o );
	}
    }

    /**
     * hit an entire list of objects
     * 
     * @param list Vector of objects to "hit"
     */

    private final void hitObjects( Vector list )
    {
	for (int i = list.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)list.elementAt( i );
	    if (o.nCount > 0)
		hitObj( o );
	}
    }

    /**
     * kill an entire list of objects
     * 
     * @param g Graphics context
     * @param list Vector of objects to kill
     */

    private final void killObjects( Graphics g, Vector list )
    {
	for (int i = list.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)list.elementAt( i );
	    if (o.nCount > 0)
	    {
		o.nCount = 0;
		killBadGuy();
		explodeObj( g, o );
		delObj( list, i );
	    }
	}
    }

    /**
     * create a letter object
     * 
     * @param cLetter character to create ($%^* create special objects)
     * @param nSize size to create (scaled by LETTER_SCALE)
     * @return created letter object
     */

    private final HyperObj newLetter( char cLetter, int nSize )
    {
	String desc = null;
	if      (cLetter >= '0' && cLetter <= '9') desc = NumberDesc[cLetter - '0'];
	else if (cLetter >= 'A' && cLetter <= 'Z') desc = LetterDesc[cLetter - 'A'];
	else if (cLetter >= 'a' && cLetter <= 'z') desc = LetterDesc[cLetter - 'a'];
	else
	{
	    int i = puncChars.indexOf( cLetter );
	    if (i >= 0) desc = PuncDesc[i];
	    else
	    {
		// handle special glyphs
		HyperObj o = null;
		switch (cLetter)
		{
		case '$': o = newObj( SpinnerShape ); break;
		case '%': o = newObj( SwarmerShape ); break;
		case '^': o = newObj( HunterShape ); break;
		case '*': o = newObj( BonusShape ); break;
		}
		if (o != null)
		{
		    o.nMass = 1;
		    o.nDir = 0;
		    o.nSpin = 5 << GAME_SCALE_BITS;
		    o.nCount = 40;
		}
		return( o );
	    }
	}
	HyperObj o = newObj( null );
	o.nMass = 1;
	o.nDir = 0;
	o.nSpin = 0;
	o.nCount = 40;
	o.nColor = 0xffffff;
	// should this be a point group instead of a polyline?
	if (desc.charAt( 0 ) == '-')
	{
	    desc = desc.substring( 1 );
	    o.nPts = -desc.length();
	}
	else o.nPts = desc.length();
	for (int i = 0; i < desc.length(); ++i)
	{
	    int j = desc.charAt( i ) - 'a';
	    // original canvas was bottom-left, so we mirror vertically to get top-left for J2ME
	    o.Pts[i][0] = (128 << GAME_SCALE_BITS) - LetterPart[j][0];
	    o.Pts[i][1] = (LetterPart[j][1] * nSize) / LETTER_SCALE;
	}
	return( o );
    }
    
    /**
     * create letter objects from a string
     *
     * @param v vector to add letters to
     * @param text text to print
     * @param Pos position to center letters about
     * @param Vel velocity of created letters
     * @param nColor color of created letters
     * @param nSize size of created letters
     */

    private final void printLetterList( Vector v, String text, int[] Pos, int[] Vel, int nColor, int nSize )
    {
	int nLen = text.length();
	int nSpace = nSize + nSize / 2;
	int nBase = (nLen - 1) * nSpace;
	int nBaseStart = Pos[0] - nBase / 2;

	for (int i = 0; i < nLen; ++i)
	{
	    HyperObj o = newLetter( text.charAt( i ), nSize / 2 );
	    if (o != null)
	    {
		v.addElement( o );
		o.Pos[0] = nBaseStart;
		o.Pos[1] = Pos[1];
		o.Vel[0] = Vel[0];
		o.Vel[1] = Vel[1];
		o.nColor = nColor;
	    }
	    nBaseStart += nSpace;
	}
    }

    /**
     * print letters to the globa LetterList
     * 
     * @param text text to print
     * @param Pos position to center letters about
     * @param Vel velocity of created letters
     * @param nColor color of created letters
     * @param nSize size of created letters
     */

    private final void printLetters( String text, int[] Pos, int[] Vel, int nColor, int nSize )
    {
	printLetterList( LetterList, text, Pos, Vel, nColor, nSize );
    }

    /**
     * create letters which spin away from the centerpoint
     * 
     * @param v vector to add letters to
     * @param text text to print
     * @param Pos position to center letters about
     * @param Vel velocity of created letters
     * @param nColor color of created letters
     * @param nSize size of created letters
     */

    private final void spinLetterList( Vector v, String text, int[] Pos, int[] Vel, int nColor, int nSize )
    {
	int nLen = text.length();
	int nSpace = nSize + nSize / 2;
	int nBase = (nLen - 1) * nSpace;
	int nBaseStart = Pos[0] - nBase / 2;

	for (int i = 0; i < nLen; ++i)
	{
	    HyperObj o = newLetter( text.charAt( i ), nSize / 2 );
	    if (o != null)
	    {
		v.addElement( o );
		int nSpin = ((i << 1) - nLen) << GAME_SCALE_BITS;
		o.Pos[0] = nBaseStart;
		o.Pos[1] = Pos[1];
		o.Vel[0] = Vel[0] + nSpin >> 2;
		o.Vel[1] = Vel[1];
		o.nColor = nColor;
		o.nSpin = nSpin;
	    }
	    nBaseStart += nSpace;
	}
    }

    /**
     * create letters which spin away from the centerpoint, add them to the global LetterList
     * 
     * @param text text to print
     * @param Pos position to center letters about
     * @param Vel velocity of created letters
     * @param nColor color of created letters
     * @param nSize size of created letters
     */

    private final void spinLetters( String text, int[] Pos, int[] Vel, int nColor, int nSize )
    {
	spinLetterList( LetterList, text, Pos, Vel, nColor, nSize );
    }

    /**
     * main update routine for all vector font letters
     * 
     * @param g Graphics context
     */

    private final void updateLetters( Graphics g )
    {
	for (int i = LetterList.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)LetterList.elementAt( i );
	    if (--o.nCount < 8)
	    {
		o.nColor = (o.nColor >> 1) & 0x7f7f7f;
		o.nColor += (o.nColor >> 1) & 0x7f7f7f;
		if (o.nCount < 1)
		    delObj( LetterList, i );
	    }
	    updateObj( g, o );
	}
    }
    
    /**
     * show the player a status message
     *
     * @param s message to be printed at the player's location
     */

    private final void printPlayerMessage( String s )
    {
	int Pos[] = { Player.Pos[0], Player.Pos[1] + 400 };
	int Vel[] = { 0, 50 };
	printLetters( s, Pos, Vel, PLAYER_MESSAGE_COLOR, 150 );
    }

    /**
     * give the player another life
     */ 

    public void addExtraLife()
    {
	printPlayerMessage( "EXTRA LIFE" );
	++Player.nCount;
	switch (++Player.nCount)
	{
	case 2: Player.nColor = PLAYER_COLOR_2; break;
	case 1: Player.nColor = PLAYER_COLOR_1; break;
	default: Player.nColor = PLAYER_COLOR; break;
	}
    }

    /**
     * accelerate the player in the specified direction
     * 
     * @param p Player object
     * @param nDir direction
     * @param nAccel amount of acceleration (delta Vel)
     */

    private final void accelPlayer( HyperObj p, int nDir, int nAccel )
    {
	nDir += p.nDir;
	accelObj( p, nDir, nAccel );
	// keep the player from moving wacky-fast
	int speed_squared = p.Vel[0] * p.Vel[0] + p.Vel[1] * p.Vel[1];
	if (speed_squared > PLAYER_SPEEDLIMIT)
	{
	    p.Vel[0] = p.Vel[0] * PLAYER_SPEEDLIMIT / speed_squared;
	    p.Vel[1] = p.Vel[1] * PLAYER_SPEEDLIMIT / speed_squared;
	}

	// special effects
	HyperObj flame = newObj( null );
	flame.Pos[0] = p.Pos[0];
	flame.Pos[1] = p.Pos[1];
	flame.Vel[0] = p.Vel[0];
	flame.Vel[1] = p.Vel[1];
	flame.nDir = nDir + ((100 + rand.nextInt( 57 )) << GAME_SCALE_BITS);
	flame.nPts = 1;
	flame.nColor = FLAME_YELLOW;
	flame.nCount = nAccel + rand.nextInt( 15 );
	accelObj( flame, flame.nDir, 50 + rand.nextInt( 20 ) );
	FlameList.addElement( flame );
	startSfx( SND_THRUST );
    }

    /**
     * fire a shot from the Player
     * 
     * @param p player object
     */

    private final void fireShot( HyperObj p )
    {
	HyperObj o = newObj( null );
	o.Pos[0] = p.Pos[0];
	o.Pos[1] = p.Pos[1];
	o.Vel[0] = p.Vel[0];
	o.Vel[1] = p.Vel[1];
	o.nMass = SHOT_MASS;
	o.nDir = p.nDir + rand.nextInt( (4 << GAME_SCALE_BITS) + 1 ) - (2 << GAME_SCALE_BITS);
	o.nSpin = 0;
	o.nCount = 16 + rand.nextInt( 8 );
	o.nColor = SHOT_COLOR;
	o.nPts = 2;
	o.Pts[0][0] = 128;
	o.Pts[0][1] = 50;
	o.Pts[1][0] = 0;
	o.Pts[1][1] = 50;
	accelObj( o, o.nDir, 100 + o.nCount );
	ShotList.addElement( o );
	startSfx( SND_PSHOT );
    }

    /**
     * kill everything off and restart at level 1
     * 
     * @param g Graphics context
     */

    private final void newGame( Graphics g )
    {
	Player.nColor = PLAYER_COLOR;
	explodeObj( g, Player );
	startSfx( SND_EXPLODE2 );
	Player.nCount = 0;
	gameLevel = 0;
	setRestart( RESTART_GAME );

	killObjects( g, SpinnerList );
	killObjects( g, SwarmerList );
	killObjects( g, HunterList );
	killObjects( g, RoidList );
    }

    /**
     * main player update routine
     * 
     * @param g Graphics context
     * @param p player object
     */

    private final void updatePlayer( Graphics g, HyperObj p )
    {
        if (p.nCount <= 0) return;
        
	// did sheilds run out?
	if (safeMillis > 0)
	{
	    if (safeMillis < frameStartMillis)
	    {
		safeMillis = 0;
		//startSfx( SND_SHEILD_OFF );
		switch (p.nCount)
		{
		case 2:
		    p.nColor = PLAYER_COLOR_2;
		    break;
		case 1:
		    p.nColor = PLAYER_COLOR_1;
		    break;
		default:
		    p.nColor = PLAYER_COLOR;
		    break;
		}
	    }
        }
	else if (key[K_SHEILD] && (gameSheilds > 0))
	{
	    --gameSheilds;
	    safeMillis = frameStartMillis + SHEILD_TIME;
	    p.nColor = PLAYER_SHEILDCOLOR;
	    //startSfx( SND_SHEILD_ON );
	    conPrint( "Player sheild" );
	}

	if (playerBombing > 0)
	{
	    if (--playerBombing <= 0)
	    {
		--gameBombs;
		killObjects( g, SpinnerList );
		killObjects( g, SwarmerList );
		killObjects( g, HunterList );
	    }
	    else
	    {
		hitObjects( SpinnerList );
		hitObjects( SwarmerList );
		hitObjects( HunterList );
	    }
	}
	else if (key[K_BOMB] && !playerBombDebounce && (gameBombs > 0))
	{
	    playerBombing = 3;
	    conPrint( "Player smartbomb" );
	}
	playerBombDebounce = key[K_BOMB];

	if (key[K_LEFT]) p.nSpin = -5 << GAME_SCALE_BITS;
	if (key[K_RIGHT]) p.nSpin = 5 << GAME_SCALE_BITS;
	if (key[K_UP]) accelPlayer( p, 128 << GAME_SCALE_BITS, 4 );
	if (key[K_DOWN]) accelPlayer( p, 0 << GAME_SCALE_BITS, 4 );
	if (playerShotDelay > 0) --playerShotDelay;
	else if (key[K_FIRE])
	{
	    playerShotDelay = PLAYER_SHOTDELAY;
	    fireShot( p );
	}

	updateObj( g, p );

	// allow the spin to decay
	p.nSpin = (p.nSpin * 9) / 10;
    }

    /**
     * main flame update routine
     * 
     * @param g Graphics context
     */

    private final void updateFlames( Graphics g )
    {
	for (int i = FlameList.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)FlameList.elementAt( i );
	    switch (--o.nCount)
	    {
	    case 11:
		o.nColor = FLAME_ORANGE;
		break;
	    case 7:
		o.nColor = FLAME_RED;
		break;
	    case 3:
		o.nColor = FLAME_DKRED;
		break;
	    case 0:
		delObj( FlameList, i );
		break;
	    }
	    updateObj( g, o );
	}
    }

    /**
     * player shot update routine, accumulates shotRect
     * 
     * @param g Graphics context
     */
    
    private final void updateShots( Graphics g )
    {
	shotRect.empty();
	for (int i = ShotList.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)ShotList.elementAt( i );
	    if (--o.nCount < 8)
	    {
		o.nColor = (o.nColor >> 1) & 0x7f7f7f;
		o.nColor += (o.nColor >> 1) & 0x7f7f7f;
		if (o.nCount < 1)
		    delObj( ShotList, i );
	    }
	    shotRect.grow( o.Pos );
	    updateObj( g, o );
	}
    }

    /**
     * handle some object hitting the player
     * 
     * @param g Graphics context
     * @param o object which hit player
     * @return true if the player was hit, false if sheilds were on or
     * player was already dead
     */

    private final boolean hitPlayer( Graphics g, HyperObj o )
    {
	if ((safeMillis > 0) || (Player.nCount <= 0))
	    return( false );

	// rumble and shake both objects
	int nMass = Player.nMass + o.nMass;
	int nSpin = Player.nSpin + o.nSpin;
	o.nSpin -= (nSpin * Player.nMass) / nMass;
	Player.nSpin -= (nSpin * o.nMass) / nMass;

	int Vel[] = { Player.Vel[0] - o.Vel[0], Player.Vel[1] - o.Vel[1] };
	o.Vel[0] += (Vel[0] * Player.nMass) / nMass;
	o.Vel[1] += (Vel[1] * Player.nMass) / nMass;
	Player.Vel[0] -= (Vel[0] * o.nMass) / nMass;
	Player.Vel[1] -= (Vel[1] * o.nMass) / nMass;

	if (--Player.nCount > 0)
	{
	    switch (Player.nCount)
	    {
	    case 2:
		Player.nColor = PLAYER_COLOR_2;
		break;
	    case 1:
		Player.nColor = PLAYER_COLOR_1;
		break;
	    default:
		Player.nColor = PLAYER_COLOR;
		break;
	    }
	    hitObj( Player );
	    startSfx( SND_PHIT );
	    return( true );
	}

	// final death
	Player.nColor = PLAYER_COLOR;
	explodeObj( g, Player );
	startSfx( SND_EXPLODE2 );
	setRestart( RESTART_ATTRACT );
	return( false );
    }

    /**
     * kill a badguy, restart next level if they're all dead
     *
     * @return true if restarting next level
     */

    private final boolean killBadGuy()
    {
	if (--gameBadGuys <= 0)
	{
	    setRestart( RESTART_NEXTLEVEL );
	    return( true );
	}
	return( false );
    }

    /**
     * create a new roid
     *
     * @param Pos position to create at
     * @param Vel base velocity
     * @param nSides number of sides for new roid
     * @param nColor color of roid
     * @param nDir direction
     * @param nSpeed speed
     * @param nSpin spin rate
     */

    private final void newRoid( int[] Pos, int[] Vel, int nSides, int nColor, int nDir, int nSpeed, int nSpin )
    {
	HyperObj o = newObj( null );

	if (nSides >= (MAX_PTS - 1))
	    nSides = MAX_PTS - 1;

	o.Pos[0] = Pos[0];
	o.Pos[1] = Pos[1];
	o.Vel[0] = Vel[0];
	o.Vel[1] = Vel[1];
	o.nMass = nSides * 128;
	o.nDir = nDir;
	o.nSpin = nSpin + rand.nextInt( (10 << GAME_SCALE_BITS) + 1 ) - (5 << GAME_SCALE_BITS);
	o.nCount = nSides * 100;
	o.nColor = nColor;
	o.nPts = nSides + 1;
	for (int i = 0; i < nSides; ++i)
	{
	    o.Pts[i][0] = i * DEG_SIZE / nSides + rand.nextInt( 30 );
	    o.Pts[i][1] = (nSides - 1) * 100 + 20 + rand.nextInt( 80 );
	}
	o.Pts[nSides][0] = o.Pts[0][0];
	o.Pts[nSides][1] = o.Pts[0][1];
	accelObj( o, nDir, nSpeed );
	RoidList.addElement( o );
	++gameBadGuys;
    }

    /**
     * a player shot hit this roid, break it up (creates new, smaller roids)
     * 
     * @param g Graphics context
     * @param o roid to break
     * @param shot shot which hit roid
     */

    private final void breakRoid( Graphics g, HyperObj o, HyperObj shot )
    {
	gameScore += o.nCount;
	if (shot != null) shot.nCount = 1;
	int nNew = 0;
	if (o.nPts > 4)
	{
	    nNew = rand.nextInt( (o.nPts - 1 ) / 2 );
	    if (o.nPts > 5) nNew += o.nPts / 4;
	}
	if (nNew == 1) // don't explode outward
	{
	    int[] Pos = { o.Pos[0], o.Pos[1] };
	    Pos[0] += rand.nextInt( 301 ) - 150;
	    Pos[1] += rand.nextInt( 301 ) - 150;
	    newRoid( Pos, o.Vel, o.nPts - 2, o.nColor, o.nDir, 8, o.nSpin );
	}
	else if (nNew > 1)
	{
	    int nSpeed = ((o.nSpin * o.nSpin * nNew) >> (GAME_SCALE_BITS << 1)) + 16;
	    for (int i = 0; i < nNew; ++i)
	    {
		int[] Pos = { o.Pos[0], o.Pos[1] };
		Pos[0] += rand.nextInt( 601 ) - 300;
		Pos[1] += rand.nextInt( 601 ) - 300;
		newRoid( Pos, o.Vel, o.nPts - (nNew + 1), o.nColor,
		    o.nDir + (((i * DEG_SIZE) << GAME_SCALE_BITS) / nNew) + rand.nextInt( 32 << GAME_SCALE_BITS ),
		    nSpeed + rand.nextInt( 1 + (gameLevel * 4) ),
		    o.nSpin / 2 );
	    }
	}
	killBadGuy();
	o.nCount = 0;
	if (nNew > 0) hitObj( o );
	else explodeObj( g, o );
	startSfx( SND_EXPLODE );
    }

    /**
     * main roid update routine
     * 
     * @param g Graphics context
     */

    private final void updateRoids( Graphics g )
    {
	for (int i = RoidList.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)RoidList.elementAt( i );
	    updateObj( g, o );
	    if (o.nCount > 0)
	    {
		rect.set( o.Pos, o.nCount );
		if (rect.hits( Player.Pos ) && hitPlayer( g, o ))
		{
		    Player.nCount = -Player.nCount;
		    Player.nColor = PLAYER_COLOR;
		    conPrint( "Player hit roid " + Player.nCount );
		    explodeObj( g, Player );
		    breakRoid( g, o, null );
		    // getting hit by an asteroid does a restart (otherwise you'd lose all
		    // your lives while you were travelling through the rock). Arguably other
		    // player hits should restart too, but that interrupts play too much.
		    if (gameBadGuys > 0) setRestart( RESTART_LEVEL );
		    else setRestart( RESTART_NEXTLEVEL );
		}
		else if (rect.hits( shotRect ))
		{
		    for (int j = ShotList.size() - 1; j >= 0; --j)
		    {
			HyperObj shot = (HyperObj)ShotList.elementAt( j );
			if (rect.hits( shot.Pos ))
			{
			    breakRoid( g, o, shot );
			    break;
			}
		    }
		}
	    }
	    else delObj( RoidList, i );
	}
    }

    /**
     * create a new bonus
     */

    private final void newBonus()
    {
	HyperObj o = newObj( BonusShape );

	o.Pos[0] = rand.nextInt( 2 * clipX + 1 ) - clipX;
	o.Pos[1] = -clipY;
	o.Vel[0] = 0;
	o.Vel[1] = 0;
	o.nMass = BONUS_MASS;
	o.nDir = rand.nextInt( DEG_SIZE << GAME_SCALE_BITS );
	o.nSpin = (rand.nextInt( 2 ) > 0) ? (12 << GAME_SCALE_BITS) : -(12 << GAME_SCALE_BITS);
	// counts map to bonus types:
	// 4: extra life (yellow)
	// 3: extra bomb (blue)
	// 2: extra sheild (green)
	// 1: points (red)
	o.nCount = rand.nextInt( 4 ) + 1;
	o.lDelay = frameStartMillis + BONUS_CHANGE_MINDELAY + rand.nextInt( BONUS_CHANGE_RANDDELAY );
	o.nColor = BONUS_COLORS[o.nCount];
	accelObj( o, o.nDir, 30 + gameLevel * 2 );
	BonusList.addElement( o );
	startSfx( SND_NEWBONUS );
    }

    /**
     * start countdown till next bonus creation
     */

    private final void setNextBonus()
    {
	nextBonusMillis = frameStartMillis + BONUS_INTERVAL;
    }

    /**
     * main bonus update routine
     * 
     * @param g Graphics context
     */

    private final void updateBonuses( Graphics g )
    {
	if (gameCreateBadGuys && (nextBonusMillis > 0) && (nextBonusMillis < frameStartMillis))
	{
	    setNextBonus();
	    newBonus();
	}

	for (int i = BonusList.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)BonusList.elementAt( i );
	    rect.set( o.Pos, BONUS_HITSIZE );
	    if (rect.hits( Player.Pos ) && (Player.nCount > 0))
	    {
		switch (o.nCount)
		{
		case 1:
		    {
			int bonus = 1000 * gameLevel + 500;
			gameScore += bonus;
			printPlayerMessage( String.valueOf( bonus ) );
		    }
		    break;
		case 2:
		    ++gameSheilds;
		    printPlayerMessage( "EXTRA SHEILD" );
		    break;
		case 3:
		    ++gameBombs;
		    printPlayerMessage( "EXTRA BOMB" );
		    break;
		case 4:
		    addExtraLife();
		    break;
		}
		o.nCount = 0;
		explodeObj( g, o );
		delObj( BonusList, i );
		startSfx( SND_BONUSGOT );
	    }
	    else if (rect.hits( shotRect ))
	    {
		for (int j = ShotList.size() - 1; j >= 0; --j)
		{
		    HyperObj shot = (HyperObj)ShotList.elementAt( j );
		    if (rect.hits( shot.Pos ))
		    {
			shot.nCount = 1;
			o.nCount = 0;
			explodeObj( g, o );
			delObj( BonusList, i );
			startSfx( SND_EXPLODE );
			break;
		    }
		}
	    }
	    else if (o.lDelay < frameStartMillis)
	    {
		if (--o.nCount > 0)
		{
		    o.nColor = BONUS_COLORS[o.nCount];
		    o.lDelay = frameStartMillis + BONUS_CHANGE_MINDELAY + rand.nextInt( BONUS_CHANGE_RANDDELAY );
		    hitObj( o );
		}
		else
		{
		    o.nCount = 0;
		    explodeObj( g, o );
		    delObj( BonusList, i );
		    startSfx( SND_EXPLODE );
		}
	    }
	    if (o.nCount > 0)
	    {
		int d = Player.Pos[0] - o.Pos[0];
		while (d < -16 || d > 16) d /= 2;
		o.Vel[0] += d - o.Vel[0] / 16;
		d = Player.Pos[1] - o.Pos[1];
		while (d < -16 || d > 16) d /= 2;
		o.Vel[1] += d - o.Vel[1] / 16;
		updateObj( g, o );
	    }
	}
    }

    /**
     * create a new spinner
     */

    private final void newSpinner()
    {
	HyperObj o = newObj( SpinnerShape );

	o.Pos[0] = rand.nextInt( 2 * clipX + 1 ) - clipX;
	o.Pos[1] = -clipY;
	o.Vel[0] = 0;
	o.Vel[1] = 0;
	o.nCount = 1 + rand.nextInt( gameLevel + 1 );
	o.nMass = SPINNER_MASS + SPINNER_MASS_DELTA * o.nCount;
	o.nDir = rand.nextInt( DEG_SIZE << GAME_SCALE_BITS );
	o.nSpin = -(12 << GAME_SCALE_BITS);
	o.nColor = SPINNER_COLOR + (o.nCount << 10);
	accelObj( o, o.nDir, 30 + gameLevel * 2 );
	SpinnerList.addElement( o );
	startSfx( SND_NEWBAD );
	++gameBadGuys;
    }

    /**
     * start countdown till next spinner creation
     */

    private final void setNextSpinner()
    {
	nextSpinnerMillis = frameStartMillis + SPINNER_INTERVAL + rand.nextInt( SPINNER_RANDINTERVAL ) - gameLevel * SPINNER_LEVEL_QUICKENER;
	if (nextSpinnerMillis < (frameStartMillis + SPINNER_MININTERVAL))
	    nextSpinnerMillis = frameStartMillis + SPINNER_MININTERVAL;
    }

    /**
     * main spinner update routine
     * 
     * @param g Graphics context
     */

    private final void updateSpinners( Graphics g )
    {
	if (gameCreateBadGuys && (nextSpinnerMillis > 0) && (nextSpinnerMillis < frameStartMillis))
	{
	    setNextSpinner();
	    if (SpinnerList.size() < MAX_SPINNERS)
		newSpinner();
	}

	for (int i = SpinnerList.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)SpinnerList.elementAt( i );
	    rect.set( o.Pos, SPINNER_HITSIZE );
	    if (rect.hits( Player.Pos ))
	    {
		hitPlayer( g, o );
		if (--o.nCount > 0)
		{
		    o.nColor = SPINNER_COLOR + (o.nCount << 10);
		    hitObj( o );
		}
		else
		{
		    killBadGuy();
		    explodeObj( g, o );
		    delObj( SpinnerList, i );
		    startSfx( SND_EXPLODE );
		}
	    }
	    else if (rect.hits( shotRect ))
	    {
		for (int j = ShotList.size() - 1; j >= 0; --j)
		{
		    HyperObj shot = (HyperObj)ShotList.elementAt( j );
		    if (rect.hits( shot.Pos ))
		    {
			shot.nCount = 1;
			gameScore += o.nCount * 500;
			if (--o.nCount > 0)
			{
			    o.nColor = SPINNER_COLOR + (o.nCount << 10);
			    hitObj( o );
			}
			else
			{
			    killBadGuy();
			    explodeObj( g, o );
			    delObj( SpinnerList, i );
			    startSfx( SND_EXPLODE );
			}
			break;
		    }
		}
	    }
	    if (o.nCount > 0)
	    {
		int d = Player.Pos[0] - o.Pos[0];
		while (d < -10 || d > 10) d /= 2;
		o.Vel[0] += d - o.Vel[0] / 8;
		d = Player.Pos[1] - o.Pos[1];
		while (d < -10 || d > 10) d /= 2;
		o.Vel[1] += d - o.Vel[1] / 8;
		updateObj( g, o );
	    }
	}
    }

    /**
     * hunter shoots something
     * 
     * @param hunter hunter firing shot
     */

    private final void fireHunterShot( HyperObj hunter )
    {
	HyperObj o = newObj( null );

	o.Pos[0] = hunter.Pos[0];
	o.Pos[1] = hunter.Pos[1];
	o.Vel[0] = hunter.Vel[0];
	o.Vel[1] = hunter.Vel[1];
	o.nMass = SHOT_MASS;
	o.nDir = hunter.nDir + rand.nextInt( (4 << GAME_SCALE_BITS) + 1 ) - (2 << GAME_SCALE_BITS);
	o.nSpin = (rand.nextInt( 2 ) > 0) ? (10 << GAME_SCALE_BITS) : (-10 << GAME_SCALE_BITS);
	o.nCount = 16 + rand.nextInt( 8 );
	o.nColor = HUNTER_SHOT_COLOR;
	o.nPts = 2;
	o.Pts[0][0] = 128;
	o.Pts[0][1] = 50;
	o.Pts[1][0] = 0;
	o.Pts[1][1] = 50;
	accelObj( o, o.nDir, 100 + o.nCount );
	HunterShotList.addElement( o );
	startSfx( SND_BSHOT );
    }

    /**
     * main update routine for shots fired by the hunters
     * 
     * @param g Graphics context
     */

    private final void updateHunterShots( Graphics g )
    {
	rect.set( Player.Pos, PLAYER_HITSIZE );
	for (int i = HunterShotList.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)HunterShotList.elementAt( i );
	    if (rect.hits( o.Pos ))
	    {
		o.nCount = 1;
		hitPlayer( g, o );
	    }
	    if (--o.nCount < 8)
	    {
		o.nColor = (o.nColor >> 1) & 0x7f7f7f;
		o.nColor += (o.nColor >> 1) & 0x7f7f7f;
		if (o.nCount < 1)
		    delObj( HunterShotList, i );
	    }
	    updateObj( g, o );
	}
    }

    /**
     * create a new hunter
     */

    private final void newHunter()
    {
	HyperObj o = newObj( HunterShape );

	o.Pos[0] = rand.nextInt( 2 * clipX + 1 ) - clipX;
	o.Pos[1] = -clipY;
	o.Vel[0] = 0;
	o.Vel[1] = 0;
	o.nMass = HUNTER_MASS;
	o.nDir = rand.nextInt( DEG_SIZE << GAME_SCALE_BITS );
	o.nSpin = 0;
	o.nCount = 1 + rand.nextInt( gameLevel + 1 );
	o.lDelay = frameStartMillis + HUNTER_CHANGE_MINDELAY + rand.nextInt( HUNTER_CHANGE_RANDDELAY );
	o.nColor = HUNTER_COLOR;
	accelObj( o, o.nDir, 30 + gameLevel * 2 );
	HunterList.addElement( o );
	startSfx( SND_NEWBAD );
	++gameBadGuys;
    }

    /**
     * start countdown till next hunter creation
     */

    private final void setNextHunter()
    {
	nextHunterMillis = frameStartMillis + HUNTER_INTERVAL + rand.nextInt( HUNTER_RANDINTERVAL ) - gameLevel * HUNTER_LEVEL_QUICKENER;
	if (nextHunterMillis < (frameStartMillis + HUNTER_MININTERVAL))
	    nextHunterMillis = frameStartMillis + HUNTER_MININTERVAL;
    }

    /**
     * main hunter update routine
     * 
     * @param g Graphics context
     */

    private final void updateHunters( Graphics g )
    {
	if (gameCreateBadGuys && (nextHunterMillis > 0) && (nextHunterMillis < frameStartMillis))
	{
	    setNextHunter();
	    if (HunterList.size() < MAX_HUNTERS)
		newHunter();
	}

	for (int i = HunterList.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)HunterList.elementAt( i );
	    rect.set( o.Pos, HUNTER_HITSIZE );
	    if (rect.hits( Player.Pos ) && hitPlayer( g, o ))
	    {
		if (--o.nCount <= 1)
		{
		    if (o.nCount == 1)
		    {
			o.nColor = HUNTER_COLOR_1;
			hitObj( o );
		    }
		    else
		    {
			killBadGuy();
			o.nColor = HUNTER_COLOR;
			explodeObj( g, o );
			delObj( HunterList, i );
			startSfx( SND_EXPLODE );
		    }
		}
	    }
	    else if (rect.hits( shotRect ))
	    {
		for (int j = ShotList.size() - 1; j >= 0; --j)
		{
		    HyperObj shot = (HyperObj)ShotList.elementAt( j );
		    if (rect.hits( shot.Pos ))
		    {
			shot.nCount = 1;
			gameScore += o.nCount * 1000;
			if (--o.nCount <= 1)
			{
			    if (o.nCount == 1)
			    {
				o.nColor = HUNTER_COLOR_1;
				hitObj( o );
			    }
			    else
			    {
				killBadGuy();
				o.nColor = HUNTER_COLOR;
				explodeObj( g, o );
				delObj( HunterList, i );
				startSfx( SND_EXPLODE );
			    }
			}
			break;
		    }
		}
	    }
	    if (o.nCount > 0)
	    {
		accelObj( o, o.nDir, 8 );
		o.Vel[0] -= o.Vel[0] / 8;
		o.Vel[1] -= o.Vel[1] / 8;
		if (o.lDelay < frameStartMillis)
		{
		    o.lDelay = frameStartMillis + HUNTER_CHANGE_MINDELAY + rand.nextInt( HUNTER_CHANGE_RANDDELAY );
		    o.nSpin = rand.nextInt( (10 << GAME_SCALE_BITS) + 1 ) - (5 << GAME_SCALE_BITS);
		    if (rand.nextInt( 4 ) > 2)
			fireHunterShot( o );
		}
		updateObj( g, o );
	    }
	}
    }

    /**
     * generate a new swarmer
     * 
     * @param Pos position to generate at
     * @param nDir direction of travel
     * @param nCount count of new swarmer (i.e. size)
     */

    private final void newSwarmer( int[] Pos, int nDir, int nCount )
    {
	HyperObj o = newObj( null );

	o.Pos[0] = Pos[0];
	o.Pos[1] = Pos[1];
	o.Vel[0] = 0;
	o.Vel[1] = 0;
	o.nMass = SWARMER_MASS + nCount;
	o.nDir = nDir;
	o.nSpin = rand.nextInt( (31 << GAME_SCALE_BITS) + 1 ) - (15 << GAME_SCALE_BITS);
	o.nCount = nCount;
	o.lDelay = frameStartMillis + SWARMER_CHANGE_MINDELAY + rand.nextInt( SWARMER_CHANGE_RANDDELAY );
	o.nColor = SWARMER_COLOR;
	o.nPts = SwarmerShape.length;
	// swarmers shrink as they split
	for (int i = 0; i < o.nPts; ++i)
	{
	    o.Pts[i][0] = SwarmerShape[i][0];
	    o.Pts[i][1] = SwarmerShape[i][1] + nCount * SWARMER_DELTASIZE;
	}
	accelObj( o, o.nDir, 30 + gameLevel * 2 );
	SwarmerList.addElement( o );
	// sound-effect differs based on how the swarmer was created
	++gameBadGuys;
    }

    /**
     * start countdown till next swarmer creation
     */

    private final void setNextSwarmer()
    {
	nextSwarmerMillis = frameStartMillis + SWARMER_INTERVAL + rand.nextInt( SWARMER_RANDINTERVAL ) - gameLevel * SWARMER_LEVEL_QUICKENER;
	if (nextSwarmerMillis < (frameStartMillis + SWARMER_MININTERVAL))
	    nextSwarmerMillis = frameStartMillis + SWARMER_MININTERVAL;
    }

    /**
     * main swarmer update routine
     * 
     * @param g Graphics context
     */

    private final void updateSwarmers( Graphics g )
    {
	if (gameCreateBadGuys && (nextSwarmerMillis > 0) && (nextSwarmerMillis < frameStartMillis))
	{
	    setNextSwarmer();
	    if (SwarmerList.size() < MAX_SWARMERS)
	    {
		int[] Pos = { rand.nextInt( clipX * 2 + 1 ) - clipX, -clipY };
		newSwarmer( Pos, rand.nextInt( DEG_SIZE << GAME_SCALE_BITS ), 8 + gameLevel * 2 );
		startSfx( SND_NEWBAD );
	    }
	}

    	for (int i = SwarmerList.size() - 1; i >= 0; --i)
	{
	    HyperObj o = (HyperObj)SwarmerList.elementAt( i );
	    rect.set( o.Pos, SWARMER_HITSIZE + o.nCount * SWARMER_DELTASIZE );
	    if (rect.hits( Player.Pos ))
	    {
		hitPlayer( g, o );
		o.nCount = 0;
	    }
	    else if (rect.hits( shotRect ))
	    {
		for (int j = ShotList.size() - 1; j >= 0; --j)
		{
		    HyperObj shot = (HyperObj)ShotList.elementAt( j );
		    if (rect.hits( shot.Pos ))
		    {
			shot.nCount = 1;
			gameScore += o.nCount * 25;
			o.nCount = 0;
			break;
		    }
		}
	    }
	    if (o.nCount > 0)
	    {
		if ((o.nCount > 1) && (o.lDelay < frameStartMillis))
		{
		    int nCount = o.nCount >> 1;
		    newSwarmer( o.Pos, o.nDir, nCount );
		    newSwarmer( o.Pos, o.nDir + (128 << GAME_SCALE_BITS), o.nCount - nCount );
		    startSfx( SND_SWARMSPLIT );
		    o.nCount = 0;
		}
		updateObj( g, o );
	    }
	    else
	    {
		killBadGuy();
		o.nColor = SWARMER_COLOR;
		explodeObj( g, o );
		delObj( SwarmerList, i );
		startSfx( SND_EXPLODE );
	    }
	}
    }

    /**
     * remove the menu from the screen
     * 
     * @param g Graphics context
     */

    private final void deleteMenu( Graphics g )
    {
	attractSwitchMillis = 0;
	for (int i = MenuList.size() - 1; i >= 0; --i)
	{
	    HyperMenuItem m = (HyperMenuItem)MenuList.elementAt( i );
	    MenuList.removeElementAt( i );
	    for (int j = m.objs.size() - 1; j >= 0; --j)
	    {
		HyperObj o = (HyperObj)m.objs.elementAt( j );
		o.nCount = 0;
		drawObj( g, o );
                //flushGraphics();
		delObj( m.objs, j );
	    }
	}
    }

    /**
     * delete any currently visible menu and create a menu from the description
     *
     * @param g Graphics context
     * @param desc array of menu description strings for HyperMenuItems
     */

    private final void createMenu( Graphics g, String[] desc )
    {
	deleteMenu( g );

	int totHeight = 0;
	for (int i = 0; i < desc.length; ++i)
	{
	    HyperMenuItem m = new HyperMenuItem( desc[i] );
	    totHeight += LETTER_SIZE * m.nSize / LETTER_SCALE;
	    MenuList.addElement( m );
	}
	totHeight += (desc.length - 1) * MENU_SPACING;

	int y = -(totHeight / 2);
	for (int i = 0; i < desc.length; ++i)
	{
	    HyperMenuItem m = (HyperMenuItem)MenuList.elementAt( i );
	    final int Pos[] = { 0, y };
	    final int Vel[] = { 0, 0 };
	    printLetterList( m.objs, m.text, Pos, Vel, m.nColor, m.nSize );
	    y += (LETTER_SIZE * m.nSize / LETTER_SCALE) + MENU_SPACING;
	}

	menuSelection = -1;
	nextMenuItem( true );
	menuDelayMillis = 0;
    }

    /**
     * start displaying the "attract" menus
     * 
     * @param g Graphics context
     */

    private final void createAttractMenu( Graphics g )
    {
	// start cycling the menus
	createMenu( g, AttractMenuDesc[gameAttractMenu] );
	attractSwitchMillis = frameStartMillis + ATTRACT_MENU_INTERVAL;
    }

    /**
     * create the options menu, updating HyperMenuItems to reflect current option settings
     * 
     * @param g Graphics context
     */

    private final void createOptionMenu( Graphics g )
    {
	OptionMenuDesc[OPT_SOUND_MENU_ENTRY] = 	"!" + String.valueOf( MENU_OPT_SOUND ) + ";sound " + (enable_snd ? "off" : "on");
	createMenu( g, OptionMenuDesc );
    }

    /**
     * deselect the current menu selection
     */
    
    private final void deselectMenu()
    {
	if (menuSelection >= 0)
	{
	    HyperMenuItem m = (HyperMenuItem)MenuList.elementAt( menuSelection );
	    for (int j = m.objs.size() - 1; j >= 0; --j)
	    {
		HyperObj o = (HyperObj)m.objs.elementAt( j );
		o.nColor = m.nColor;
	    }
	}
    }

    /**
     * select the menu corresponding to the command
     *
     * @param nCmd command item to select
     * @return selected item, or -1 if not found in current menu
     */
    
    private final int selectMenu( int nCmd )
    {
	deselectMenu();

	for (int i = 0; i < MenuList.size(); ++i)
	{
	    HyperMenuItem m = (HyperMenuItem)MenuList.elementAt( i );
	    if (m.nCmd == nCmd) return( menuSelection = i );
	}
	return( menuSelection = -1 );
    }

    /**
     * advance selection to next or previous menu item which corresponds to a command
     * 
     * @param forward direction to move selection, true means next item
     * @return menu selection or -1 if no selectable items exist in this menu
     */

    private final int nextMenuItem( boolean forward )
    {
	deselectMenu();

	int dir = forward ? 1 : -1;
	for (int i = 0; i < MenuList.size(); ++i)
	{
	    menuSelection = (menuSelection + dir) % MenuList.size();
	    if (menuSelection < 0) menuSelection += MenuList.size();
	    HyperMenuItem m = (HyperMenuItem)MenuList.elementAt( menuSelection );
	    if (m.nCmd > 0) return( menuSelection );
	}
	return( menuSelection = -1 );
    }

    /**
     * move the high score selection point within the special high score menu item
     * 
     * @param m menu item corresponding to high score input
     * @param forward direction to advance character selection, true means next char
     */

    private final void nextHighScoreChar( HyperMenuItem m, boolean forward )
    {
	int dir = forward ? 1 : -1;
	char c = lastHighScoreName[menuHighScoreChar];
	int i = (highScoreChars.lastIndexOf( c ) + dir) % highScoreChars.length();
	if (i < 0) i += highScoreChars.length();
	c = highScoreChars.charAt( i );
	lastHighScoreName[menuHighScoreChar] = c;
	HyperObj o = (HyperObj)m.objs.elementAt( menuHighScoreChar );
	HyperObj n = newLetter( c, m.nSize / 2 );
	n.Pos[0] = o.Pos[0];
	n.Pos[1] = o.Pos[1];
	n.Vel[0] = o.Vel[0];
	n.Vel[1] = o.Vel[1];
	n.nColor = o.nColor;
	m.objs.setElementAt( n, menuHighScoreChar );
	o.nCount = 10;
	o.Vel[1] = dir << GAME_SCALE_BITS;
	LetterList.addElement( o );
    }

    /**
     * set the highscore and user's name in the permanent store
     * 
     * @param g Graphics context
     */

    private final void setHighScore( Graphics g )
    {
	HighScore hs = new HighScore( String.valueOf( lastHighScoreName ), gameScore );
	for (int i = gameHighScore.length - 1; i >= 0; --i)
	{
	    if (gameScore > gameHighScore[i].score)
	    {
		if (i < gameHighScore.length - 1)
		    gameHighScore[i + 1] = gameHighScore[i];
		gameHighScore[i] = hs;
	    }
	}

	prefs.set( PREFKEY_LASTHSNAME, String.valueOf( lastHighScoreName ) );
	for (int i = gameHighScore.length - 1; i >= 0; --i)
	{
	    AttractMenuHighScoreDesc[FIRST_HIGHSCORE_MENU_ENTRY + i] = gameHighScore[i].getMenuText();
	    prefs.set( PREFKEY_HS + i, gameHighScore[i].getText() );
	    conPrint( PREFKEY_HS + i + "=" + gameHighScore[i].getText() );
	}
	prefs.sync();

	gameSetHighScore = false;
	gameAttractMenu = ATTRACT_MENU_HIGHSCORE;
	createAttractMenu( g );
    }

    /**
     * interrupt what we're doing and show a root for the menu system on next draw
     */

    public final void showRootMenu()
    {
	if (gameMode != APP_MENU)
	{
	    attractSwitchMillis = 0;
	    menuModeStartMillis = frameStartMillis;
	    menuCreateRoot = true;
	    menuResumeMode = gameMode;
	    gameMode = APP_MENU;
	}
    }

    /**
     * continue doing whatever we were doing before the root menu was shown
     *
     * @param g graphics context
     */
    
    private void hideMenu( Graphics g )
    {
	if (gameMode == APP_MENU)
	{
	    long pausedMillis = frameStartMillis - menuModeStartMillis;
	    if (restartMillis > 0) restartMillis += pausedMillis;
	    if (safeMillis > 0) safeMillis += pausedMillis;
	    if (nextBonusMillis > 0) nextBonusMillis += pausedMillis;
	    if (nextHunterMillis > 0) nextHunterMillis += pausedMillis;
	    if (nextSwarmerMillis > 0) nextSwarmerMillis += pausedMillis;
	    if (nextSpinnerMillis > 0) nextSpinnerMillis += pausedMillis;
	    deleteMenu( g );
	    gameMode = menuResumeMode;
	    if (gameMode == APP_ATTRACT)
		createAttractMenu( g );
	}
    }

    /**
     * handle menu input and draw menus
     * 
     * @param g Graphics context
     */

    private final void updateMenu( Graphics g )
    {
	if ((attractSwitchMillis > 0) && (attractSwitchMillis < frameStartMillis))
	{
	    gameAttractMenu = (gameAttractMenu + 1) % NUM_ATTRACT_MENUS;
	    createAttractMenu( g );
	}

	// animate the current selection
	if (menuSelection >= 0)
	{
	    HyperMenuItem m = (HyperMenuItem)MenuList.elementAt( menuSelection );
	    int red = (m.nColor >> 16) & 0xff;
	    int grn = (m.nColor >>  8) & 0xff;
	    int blu = (m.nColor >>  0) & 0xff;
	    int nCos = (fxCos[((int)frameStartMillis >> 1) & DEG_MASK] + 0x10000) >> 1;
	    red += ((0xff - red) * nCos) >> 16;
	    grn += ((0xff - grn) * nCos) >> 16;
	    blu += ((0xff - blu) * nCos) >> 16;
	    int nColor = (red << 16) | (grn << 8) | blu;
	    switch (m.nCmd)
	    {
	    case MENU_SET_HIGHSCORE:
		for (int j = 0; j < 3; ++j)
		{
		    HyperObj o = (HyperObj)m.objs.elementAt( j );
		    o.nColor = (menuHighScoreChar == j) ? nColor : m.nColor;
		}
		break;
	    default:
		for (int j = m.objs.size() - 1; j >= 0; --j)
		{
		    HyperObj o = (HyperObj)m.objs.elementAt( j );
		    o.nColor = nColor;
		}
		break;
	    }
	}

	for (int i = MenuList.size() - 1; i >= 0; --i)
	{
	    HyperMenuItem m = (HyperMenuItem)MenuList.elementAt( i );
	    for (int j = m.objs.size() - 1; j >= 0; --j)
	    {
		HyperObj o = (HyperObj)m.objs.elementAt( j );
		updateObj( g, o );
	    }
	}

	if (menuSelection >= 0)
	{
	    // time to poll the keyboard yet?
	    if (!menuFireKeyDebounce && (last_keystate == 0))
	    {
		menuDelayMillis = 0;
		return;
	    }
	    if (menuDelayMillis > frameStartMillis) return;

	    HyperMenuItem m = (HyperMenuItem)MenuList.elementAt( menuSelection );

	    switch (m.nCmd)
	    {
	    case MENU_SET_HIGHSCORE:
		if (key[K_UP] || key[K_DOWN])
		{
		    nextHighScoreChar( m, key[K_DOWN] );
		    startSfx( SND_EXPLODE );
		}
		if (menuFireKeyDebounce && !key[K_FIRE])
		{
		    if (++menuHighScoreChar >= 3)
		    {
			hitObjects( m.objs );
			startSfx( SND_BONUSGOT );
			setHighScore( g );
		    }
		    else startSfx( SND_PHIT );
		}
		if (key[K_LEFT] && (menuHighScoreChar > 0))
		{
		    --menuHighScoreChar;
		    startSfx( SND_PHIT );
		}
		if (key[K_RIGHT] && (menuHighScoreChar < 2))
		{
		    ++menuHighScoreChar;
		    startSfx( SND_PHIT );
		}
		break;
	    default:
		if (key[K_UP] || key[K_DOWN])
		{
		    nextMenuItem( key[K_DOWN] );
		    startSfx( SND_EXPLODE );
		}
		if (menuFireKeyDebounce && !key[K_FIRE])
		{
		    hitObjects( m.objs );
		    startSfx( SND_BONUSGOT );
		    switch (m.nCmd)
		    {
		    case MENU_RESUME:
			hideMenu( g );
			break;
		    case MENU_NEWGAME:
			hideMenu( g );
			newGame( g );
			break;
		    case MENU_BACK:
			createMenu( g, RootMenuDesc );
			break;
		    case MENU_EXIT:
			conPrint( "Exiting game..." );
			stop();
			break;
		    case MENU_ABOUT:
			createMenu( g, AboutMenuDesc );
			break;
		    case MENU_LICENSE:
			licenseLine = 0;
			licenseTicks = LICENSE_DELAY;
			hideMenu( g );
			break;
		    case MENU_HELP:
			createMenu( g, HelpMenuDesc );
			break;
		    case MENU_OPTIONS:
			createOptionMenu( g );
			break;
		    case MENU_OPT_SOUND:
			enable_snd ^= true;
			prefs.set( PREFKEY_SOUND, String.valueOf( enable_snd ) );
			prefs.sync();
			createOptionMenu( g );
			selectMenu( MENU_OPT_SOUND );
			break;
		    case MENU_OPT_RESET:
			loadPreferences( true );
			hideMenu( g );
			break;
		    }
		}
		break;
	    }
	    
	    // make sure we don't fall out of the menu and into something unintended
	    menuFireKeyDebounce = key[K_FIRE];
	    menuDelayMillis = frameStartMillis + MENU_DELAY;
	}
    }

    /**
     * restart a game after a delay
     * 
     * @param mode one of
     *   RESTART_ATTRACT: go into APP_ATTRACT mode, show "GAME OVER" if currently in APP_PLAY mode
     *   RESTART_GAME: switch to APP_PLAY mode
     *   RESTART_LEVEL: show "GET READY" and restart level
     *   RESTART_NEXTLEVEL: show "LEVEL N" and start next level
     */

    private final void setRestart( int mode )
    {
	// is a restart already in progress?
	if (restartMillis > 0) return;

	// no mode badguys should pop up till restartGame
	gameCreateBadGuys = false;

	// make sure new badguys don't pop up while waiting for restart
	nextBonusMillis = 0;
	nextHunterMillis = 0;
	nextSwarmerMillis = 0;
	nextSpinnerMillis = 0;

	final int Pt[] = { 0, 0 };
	switch (mode)
	{
	case RESTART_ATTRACT:
	    if (gameMode == APP_PLAY)
	    {
		spinLetters( "GAME OVER", Pt, Pt, TEXT_RED, 200 );
		gameSetHighScore = (gameScore > gameHighScore[MAX_HIGHSCORES - 1].score);
	    }
	    gameMode = APP_ATTRACT;
	    restartMillis = frameStartMillis + RESTART_ATTRACT_DELAY;
	    break;
	case RESTART_GAME:
	    gameMode = APP_PLAY;
	    restartMillis = frameStartMillis + RESTART_GAME_DELAY;
	    break;
	case RESTART_LEVEL:
	    printLetters( "GET READY", Pt, Pt, TEXT_BLUE, 150 );
	    restartMillis = frameStartMillis + RESTART_LEVEL_DELAY;
	    break;
	case RESTART_NEXTLEVEL:
	    printLetters( "LEVEL " + (gameLevel + 1), Pt, Pt, TEXT_BLUE, 150 );
	    restartMillis = frameStartMillis + RESTART_NEXTLEVEL_DELAY;
	    break;
	}
    }

    /**
     * start up a new game when restart timer expires
     * 
     * @param gfx Graphics context
     */

    private final void restartGame( Graphics gfx )
    {
	conPrint( "restartGame()" );

	// reset all timers and counters
	restartMillis = 0;
	setNextBonus();
	setNextHunter();
	setNextSwarmer();
	setNextSpinner();
	playerShotDelay = 0;
	playerBombing = 0;

	switch (gameMode)
	{
	case APP_ATTRACT:
	    Player.nCount = 0;
	    Player.Pos[0] = 0;
	    Player.Pos[1] = 0;
	    Player.Vel[0] = 0;
	    Player.Vel[1] = 0;
	    gameAttractMenu = 0;
	    if (gameSetHighScore)
	    {
		HighScoreMenuDesc[HIGHSCORE_MENU_ENTRY] = "!" + Integer.toString( MENU_SET_HIGHSCORE ) + ";@250;" + String.valueOf( lastHighScoreName ) + " - " + gameScore;
		menuHighScoreChar = 0;
		createMenu( gfx, HighScoreMenuDesc );
	    }
	    else
	    {
		createAttractMenu( gfx );
	    }
	    gameCreateBadGuys = false;
	    break;

	case APP_PLAY:
	    deleteMenu( gfx );
	    gameCreateBadGuys = true;
	    if (Player.nCount == 0)
	    {
		int[] Pos = { 0, clipY / 2 };
		int[] Vel = { 0, -100 };
		printLetters( "HYPEROIDS", Pos, Vel, TEXT_YELLOW, 300 );
		startSfx( SND_TITLE );

		Player.nCount = 3;
		gameLastLifeScore = gameScore = 0;
		// keep playing from the current level...
		//gameLevel = 0;
		gameSheilds = gameBombs = 3;
	    }
	    else if (Player.nCount < 0)
	    {
		// cheesy way of restarting after a major collision
		Player.nCount = -Player.nCount;
		gameSheilds = gameBombs = 3;
	    }

	    Player.Pos[0] = Player.Pos[1] = 0;
	    Player.Vel[0] = Player.Vel[1] = 0;
	    // pointing up
	    Player.nDir = 192 << GAME_SCALE_BITS;
	    Player.nSpin = 0;
	    Player.nColor = PLAYER_SHEILDCOLOR;
	    // start with a free sheild powerup
	    safeMillis = frameStartMillis + SHEILD_TIME * 2;
	    break;
	}

	// re-seed the asteroid field even when gameCreateBadGuys == false
	// (asteroids aren't really badguys, they're just misunderstood)
	if (gameBadGuys == 0)
	{
	    ++gameLevel;
	    for (int nCnt = 5 + gameLevel; nCnt > 0; --nCnt)
	    {
		int[] Pos = { rand.nextInt( clipX * 2 + 1 ) - clipX, rand.nextInt( clipY * 2 + 1 ) - clipY };
		int[] Vel = { 0, 0 };
		// avoid creating roids in the center
		while ((Pos[0] * Pos[0] + Pos[1] * Pos[1]) < (SAFE_ZONE_SIZE << GAME_SCALE_BITS))
		{
		    Pos[0] = rand.nextInt( clipX * 2 + 1 ) - clipX;
		    Pos[1] = rand.nextInt( clipY * 2 + 1 ) - clipY;
		}
		// pick a random greyish/brownish color
		int b = rand.nextInt( 64 ) + 64;
		int g = b + rand.nextInt( 32 );
		int r = g + rand.nextInt( 32 );
		int c = (r << 16) | (g << 8) | b;
		newRoid( Pos, Vel, 6 + rand.nextInt( 2 ), c, rand.nextInt( DEG_SIZE << GAME_SCALE_BITS ), 15 + rand.nextInt( gameLevel * 8 + 1 ), 0 );
	    }
	}
    }

    private FileConnection fcLog = null;
    private DataOutputStream osLog = null;

    /**
     * open a logfile to dump debugging information to
     */

    private final void openLogfile()
    {
	try
	{
	    fcLog = (FileConnection)Connector.open( "file:///e:/hyperoid.log" );
	    // If no exception is thrown, then the URI is valid, but the file may or may not exist.
	    if (fcLog.exists())
		fcLog.delete();
	    fcLog.create();
	    osLog = fcLog.openDataOutputStream();
	}
	catch (IOException ioe)
	{
	    System.out.println( "Failed to open logfile: " + ioe.toString() );
	}
    }

    /**
     * close the logfile if opened
     */

    private final void closeLogfile()
    {
	try
	{
	    if (osLog != null)
	    {
		osLog.flush();
		osLog.close();
		osLog = null;
	    }
	    if (fcLog != null)
	    {
		fcLog.close();
		fcLog = null;
	    }
	}
	catch (IOException ioe)
	{
	    System.out.println( "Failed to close logfile: " + ioe.toString() );
	}
    }

    /**
     * print a string to the logfile if opened
     * 
     * @param s string to print
     */

    private final void printLog( String s )
    {
	if (osLog != null)
	{
	    try
	    {
		// de-unicodify
		s += "\n";
		osLog.write( s.getBytes( "UTF8" ) );
	    }
	    catch (IOException ioe)
	    {
		System.out.println( "Failed to write \"" + s + "\" to logfile: " + ioe.toString() );
	    }
	}
    }

    /**
     * When fullscreen mode is set, some devices will call
     * this method to notify us of the new width/height.
     * 
     * @param newWidth new width resized to
     * @param newHeight new height resized to 
     */

    public void sizeChanged( int newWidth, int newHeight )
    {
        conPrint( "Resized to (" + newWidth + "," + newHeight + ")" );
    }

    /**
     * drawGlyph
     * 
     * @param g Graphics context
     * @param id IMG_* glyph to draw
     * @param xpos glyph x position to draw at (scaled by IMGSIZE_SIZE)
     * @param ypos glyph y position to draw at (scaled by IMGSIZE_SIZE)
     */

    private final void drawGlyph( Graphics g, int id, int xpos, int ypos )
    {
        g.drawImage( image[id], xpos << 4, ypos << 4, Graphics.TOP | Graphics.LEFT );
    }

    /**
     * draw a numeric value as digits from the set of images
     * 
     * @param g Graphics context
     * @param value numeric value to display
     * @param numdigits number of decimal places to draw
     * @param xpos glyph x position to draw at (scaled by IMGSIZE_SIZE)
     * @param ypos glyph y position to draw at (scaled by IMGSIZE_SIZE)
     */

    private final void drawNumGlyphs( Graphics g, int value, int numdigits, int xpos, int ypos )
    {
        for (int i = 0; i < numdigits; ++i)
        {
            drawGlyph( g, IMG_0 + (value % 10), xpos + numdigits - 1 - i, ypos );
            value /= 10;
        }
    }

    /**
     * draw a row of icons up to numicons long, or numicons - 1 + "+"
     * 
     * @param g Graphics context
     * @param id IMG_* glyph to draw
     * @param align alignment, 0 => left, 1 => center, 2 => right alignment
     * @param value quantity of glyphs to draw
     * @param numicons maximum number of glyphs to draw
     * @param xpos glyph x position to draw at (scaled by IMGSIZE_SIZE)
     * @param ypos glyph y position to draw at (scaled by IMGSIZE_SIZE)
     */

    private final void drawIconRow( Graphics g, int id, int align, int value, int numicons, int xpos, int ypos )
    {
        if (value > numicons)
        {
            for (int i = 0; i < numicons - 1; ++i)
                drawGlyph( g, id, xpos + i, ypos );
            drawGlyph( g, IMG_PLUS, xpos + numicons - 1, ypos );
        }
        else
	{
	    // 0 => left, 1 => center, 2 => right alignment
	    align = (align * ((numicons - value) << 4)) >> 1;
	    g.translate( align, 0 );
	    for (int i = 0; i < value; ++i)
	    {
		drawGlyph( g, id, xpos + i, ypos );
	    }
	    g.translate( -align, 0 );
	}
    }

    // console support - a debugging output
    private final Vector ConsoleStrings = new Vector();
    private int consoleLines = 0;
    private int consoleTimeout = 0;

    /**
     * optionally prints information to the screen and/or a log file
     * 
     * @param s text to print
     */

    public final void conPrint( String s )
    {
	if (!running)
	{
	    ConsoleStrings.addElement( s );
	    // render thread not running, update display ourselves during startup
	    Graphics g = getGraphics();
	    final int w = getWidth();
	    final int h = getHeight();
	    g.setColor( 0x000000 );
	    g.fillRect( 0, 0, w, h );
	    Font f = g.getFont();
	    final int line_h = f.getHeight() + 2;
	    final int lines = h / line_h;
	    int y = h - line_h;
	    g.translate( -g.getTranslateX(), -g.getTranslateY() );
	    g.setColor( 0xff7f7f );
	    for (int i = ConsoleStrings.size() - 1; (i >= 0) && (y >= 0); --i, y -= line_h)
	    {
		g.drawString( (String)ConsoleStrings.elementAt( i ), 3, y, Graphics.TOP | Graphics.LEFT );
	    }
	    flushGraphics();
	}
	else if (console)
	{
	    ConsoleStrings.addElement( s );
	    while (ConsoleStrings.size() > 80)
		ConsoleStrings.removeElementAt( 0 );
	    clearScreen = true;
	    ++consoleLines;
	    consoleTimeout = 120;
	}
	if (logging)
	    printLog( s );
    }

    /**
     * main routine to update and draw everything, called from run thread
     * 
     * @param g Graphics context
     */

    private final void draw( Graphics g )
    {
        final int w = getWidth();
        final int h = getHeight();

	// calculate playfield clipping region (+/- values around center)
        clipX = (w >> 1) << GAME_SCALE_BITS;
        clipY = ((h - (2 * IMGSIZE_SIZE)) >> 1) << GAME_SCALE_BITS;

	if ((gameScore - gameLastLifeScore) > EXTRA_LIFE_SCORE)
	{
	    gameLastLifeScore += EXTRA_LIFE_SCORE;
	    addExtraLife();
	}
	gameLives = (Player.nCount > 0) ? Player.nCount : -Player.nCount;

	if (clearScreen || (gameScore != displayedScore) || (gameLevel != displayedLevel))
        {
            displayedLevel = gameLevel;
            displayedScore = gameScore;
            g.translate( -g.getTranslateX(), -g.getTranslateY() );
            g.setClip( 0, 0, w, IMGSIZE_SIZE );
            g.setColor( 0xc0c0c0 );
            g.fillRect( 0, 0, w, IMGSIZE_SIZE );
	    int spots = w >> IMGSIZE_BITS;
            int score_spots = spots - 6;
            drawGlyph( g, IMG_SCORE, 0, 0 );
            drawNumGlyphs( g, gameScore, score_spots, 1, 0 );
            drawGlyph( g, IMG_LEVEL, spots - 4, 0 );
            drawNumGlyphs( g, gameLevel, 3, spots - 3, 0 );
        }

        if (clearScreen || (gameSheilds != displayedSheilds) || (gameLives != displayedLives) || (gameBombs != displayedBombs))
        {
            displayedSheilds = gameSheilds;
            displayedLives = gameLives;
            displayedBombs = gameBombs;
            g.translate( -g.getTranslateX(), h - IMGSIZE_SIZE - g.getTranslateY() );
            g.setClip( 0, 0, w, IMGSIZE_SIZE );
            g.setColor( 0xc0c0c0 );
            g.fillRect( 0, 0, w, IMGSIZE_SIZE );
	    int spots = w >> IMGSIZE_BITS;
            int powerup_spots = (spots - 2) / 3;
	    int life_spots = (w >> IMGSIZE_BITS) - ((powerup_spots << 1) + 2);
            drawIconRow( g, IMG_SHEILD, 0, gameSheilds, powerup_spots, 0, 0 );
            drawIconRow( g, IMG_LIFE, 1, gameLives, life_spots, powerup_spots + 1, 0 );
            drawIconRow( g, IMG_BOMB, 2, gameBombs, powerup_spots, spots - powerup_spots, 0 );
        }

        // setup the playfield and clear if requested
        g.translate( -g.getTranslateX(), -g.getTranslateY() );
        g.setClip( 0, IMGSIZE_SIZE, w, h - (2 * IMGSIZE_SIZE) );
        if (clearScreen)
        {
            g.setColor( 0x000000 );
            g.fillRect( 0, IMGSIZE_SIZE, w, h - (2 * IMGSIZE_SIZE) );
        }

	// put the playfield origin in the screen's center
        g.translate( w >> 1, h >> 1 );

	// check for game restarts (needs to be here so graphics is setup correctly)
	if ((gameMode != APP_MENU) && (restartMillis > 0) && (restartMillis < frameStartMillis))
	    restartGame( g );

	// update and redraw everything
	switch (gameMode)
	{
	default:
	case APP_ATTRACT:
	    if (!gameSetHighScore && (last_keystate != 0))
		setRestart( RESTART_GAME );
	    // fall through...
	case APP_PLAY:
	    updatePlayer( g, Player );
	    updateFlames( g );
	    updateShots( g );
	    updateRoids( g );
	    updateSpinners( g );
	    updateSwarmers( g );
	    updateHunters( g );
	    updateHunterShots( g );
	    updateLetters( g );
	    updateBonuses( g );
	    break;
	case APP_MENU:
	    if (menuCreateRoot)
	    {
		menuCreateRoot = false;
		createMenu( g, RootMenuDesc );
	    }
	    drawObj( g, Player );
	    drawObjects( g, RoidList );
	    drawObjects( g, ShotList );
	    updateFlames( g );
	    drawObjects( g, SpinnerList );
	    drawObjects( g, HunterList );
	    drawObjects( g, HunterShotList );
	    drawObjects( g, SwarmerList );
	    updateLetters( g );
	    drawObjects( g, BonusList );
	    break;
	}
	if (MenuList.size() > 0)
	    updateMenu( g );

	if ((licenseTicks > 0) && (--licenseTicks == 0))
	{
	    final int[] Pos = { 0, clipY };
	    final int[] Vel = { 0, -40 };
	    int i = LetterList.size();
	    printLetters( LICENSE[licenseLine], Pos, Vel, TEXT_BLUE, 128 );
	    while (i < LetterList.size())
	    {
		HyperObj o = (HyperObj)LetterList.elementAt( i++ );
		// if this is set too high, we tend to exhaust the java heap (at least on the emulator)
		o.nCount = 60;
	    }
	    if (++licenseLine >= LICENSE.length)
	    {
		licenseLine = 0;
		licenseTicks = 0;
	    }
	    else licenseTicks = LICENSE_DELAY;
	}

	if (draw_keys)
	{
	    g.translate( -g.getTranslateX(), -g.getTranslateY() );
	    String keys = "0000000" + Integer.toHexString( last_keystate );
	    keys = keys.substring( keys.length() - 8 );
            Font f = g.getFont();
            g.setColor( 0x000000 );
            g.fillRect( 4, 19, f.stringWidth( keys ) + 2, f.getHeight() + 2 );
	    g.setColor( 0xffffff );
	    g.drawString( keys, 5, 20, Graphics.TOP | Graphics.LEFT );
	}

	if (draw_fps && (elapsedMillis > 0))
        {
            g.translate( -g.getTranslateX(), -g.getTranslateY() );
            String fps = "fps: " + (int)(1000.0f / elapsedMillis);
            Font f = g.getFont();
            g.setColor( 0x000000 );
            g.fillRect( 4, 19, f.stringWidth( fps ) + 20, f.getHeight() + 2 );
            g.setColor( 0xff0000 );
            g.drawString( fps, 5, 20, Graphics.TOP | Graphics.LEFT );
        }

	if (draw_cont_types)
	{
	    String[] cont_types = Manager.getSupportedContentTypes( null );
            g.translate( -g.getTranslateX(), -g.getTranslateY() );
            g.setColor( 0xffffff );
	    for (int i = 0; i < cont_types.length; ++i)
		g.drawString( cont_types[i], 5, 20 + i * 15, Graphics.TOP | Graphics.LEFT );
	}

	if (consoleTimeout > 0)
	{
	    --consoleTimeout;
	    Font f = g.getFont();
	    final int line_h = f.getHeight() + 2;
	    final int lines = (h - 32) / line_h;
	    int y = h - 16 - line_h;
            g.translate( -g.getTranslateX(), -g.getTranslateY() );
	    if (consoleTimeout < 16)
	    {
		int r = (consoleTimeout << 4) | consoleTimeout;
		int gb = r >> 1;
		g.setColor( (r << 16) | (gb << 8) | gb );
	    }
	    else g.setColor( 0xff7f7f );
	    int shown = 0;
	    for (int i = ConsoleStrings.size() - 1; (i >= 0) && (y >= 16) && (shown < consoleLines); --i, y -= line_h, ++shown)
	    {
		g.drawString( (String)ConsoleStrings.elementAt( i ), 3, y, Graphics.TOP | Graphics.LEFT );
	    }
	}
	else
	{
	    consoleLines = 0;
	    clearScreen = false;
	}

	flushGraphics();
    }

    /**
     * stupid hack to keep backlight on on my SonyEricsson w900i
     */

    private final void checkBacklight()
    {
        if ((frameStartMillis - lastBacklightMillis) > 10000)
        {
            lastBacklightMillis = frameStartMillis;
            dpy.flashBacklight( 1 );
            dpy.flashBacklight( 0 );
        }
    }

    /**
     * start a one-shot sound effect
     * 
     * @param id SND_* effect to start
     */
    
    private final void startSfx( int id )
    {
	if (!enable_snd)
	    return;
	if ((toneEndMillis > 0) && ((toneEndMillis - frameStartMillis) > MAX_TONE_QUEUE))
	    return;

	try
	{
	    switch (id)
	    {
	    case SND_BONUSGOT:
	    case SND_NEWBAD:
	    case SND_NEWBONUS:
	    case SND_PHIT:
	    case SND_TITLE:
	    case SND_SHEILD_ON:
	    case SND_SHEILD_OFF:
		if (sound[id].getState() != javax.microedition.media.Player.STARTED)
		    sound[id].start();
		break;
	    case SND_EXPLODE2:
		for (int i = 0; i < 5; ++i)
		    Manager.playTone( 35 + rand.nextInt( 10 ), 2, 100 );
		toneEndMillis += ((toneEndMillis <= 0) ? frameStartMillis : 0) + 10;
		// fall through...
	    case SND_EXPLODE:
		for (int i = 0; i < 5; ++i)
		    Manager.playTone( 30 + rand.nextInt( 10 ), 2, 100 );
		toneEndMillis += ((toneEndMillis <= 0) ? frameStartMillis : 0) + 10;
		break;
	    case SND_SWARMSPLIT:
		for (int i = 0; i < 10; i += 2)
		    Manager.playTone( 50 - i, 2, 100 );
		toneEndMillis += ((toneEndMillis <= 0) ? frameStartMillis : 0) + 10;
		break;
	    case SND_PSHOT:
	    case SND_THRUST:
	    case SND_BSHOT:
	    case SND_BWOUND:
		// this actually works! (much lower latency than player.start()) on the w900i
		{
		    int ms = ((int)elapsedMillis / 3) + 1;
		    Manager.playTone( 69 + id * 5, ms, 100 );
		    toneEndMillis += ((toneEndMillis <= 0) ? frameStartMillis : 0) + ms;
		}
		break;
	    }
	}
	catch (MediaException me)
	{
	    conPrint( "startSfx exception " + me.toString() );
	}
    }

    /**
     * Starts the canvas by firing up a thread
     *
     * @throws java.lang.Error an error of some kind happened
     */

    public void start() throws Error
    {
        conPrint( "Canvas (" + getWidth() + "x" + getHeight() + ")" );

	// load the image resources
        image = new Image[imgnames.length];
        try
        {
            for (int i = 0; i < imgnames.length; ++i)
            {
                conPrint( "Loading " + imgnames[i] );
                image[i] = Image.createImage( imgnames[i] );
            }
        }
        catch (IOException io)
        {
            throw new Error( "Failed to load image" );
        }

	// load the sound effects
	try
	{
	    // we have samples for everything...
	    // ...but latency sucks so badly we have to use the hw tone generator in some cases
	    sound = new Player[sndnames.length];
	    for (int i = 0; i < sndnames.length; ++i)
	    {
		conPrint( "Loading " + sndnames[i] );
		InputStream is = getClass().getResourceAsStream( sndnames[i] );
		sound[i] = Manager.createPlayer( is, "audio/x-wav" );
		sound[i].setLoopCount( 1 );
		sound[i].prefetch();
	    }
	}
	catch (IOException ioe)
	{
	    conPrint( "Audio: " + ioe.toString() );
            //throw new Error( "Failed to load sound" );
	}
	catch (MediaException me)
	{
	    conPrint( "Audio: " + me.toString() );
            //throw new Error( "Failed to load sound" );
	}

        // only need to create the player once
        conPrint( "Creating player" );
        Player = newObj( PlayerShape );
        Player.nMass = PLAYER_MASS;
        Player.nCount = 0;
        final int w = getWidth();
        final int h = getHeight();
        clipX = (w >> 1) << GAME_SCALE_BITS;
        clipY = ((h - 32) >> 1) << GAME_SCALE_BITS;

	// Make sure we know we are running
        running = true;
        done = false;
	// make sure all timers are set relative to current time
        frameStartMillis = System.currentTimeMillis();
	// force an immediate restart into attract mode
	restartMillis = 1;
	gameMode = APP_ATTRACT;

	if (run_as_thread)
	{
	    // seems to have less stutter to run as a thread rather than inline
	    conPrint( "Creating game thread..." );
	    Thread myThread = new Thread( this );
	    myThread.setPriority( Thread.MAX_PRIORITY );
	    myThread.start();
	    conPrint( "Running " + myThread.toString() );
	}
	else
	{
	    run();
	}
    }

    /**
     * runs the whole thread. Also keeps track of FPS. Cleans up resources on exit.
     */

    public void run()
    {
        System.gc();
	conPrint( "Main loop running..." );

	while (running)
        {
            try
            {
		if (toneEndMillis < frameStartMillis)
		    toneEndMillis = 0;

		last_keystate = getKeyStates();
		key[K_FIRE]	= ((last_keystate & GameCanvas.FIRE_PRESSED) != 0);
		key[K_UP]	= ((last_keystate & GameCanvas.UP_PRESSED) != 0);
		key[K_DOWN]	= ((last_keystate & GameCanvas.DOWN_PRESSED) != 0);
		key[K_LEFT]	= ((last_keystate & GameCanvas.LEFT_PRESSED) != 0);
		key[K_RIGHT]	= ((last_keystate & GameCanvas.RIGHT_PRESSED) != 0);
		key[K_SHEILD]	= ((last_keystate & GameCanvas.GAME_A_PRESSED) != 0);
		key[K_BOMB]	= ((last_keystate & GameCanvas.GAME_B_PRESSED) != 0);

		draw( getGraphics() );
                
                checkBacklight();

                // Sleep to prevent starvation
                long currentMillis = System.currentTimeMillis();
                elapsedMillis = currentMillis - frameStartMillis;
                frameStartMillis = currentMillis;
                long sleepMillis = fpsMillis - elapsedMillis;
                if (sleepMillis > 0)
		{
		    //conPrint( "Elapsed " + elapsedMillis + " sleeping " + sleepMillis );
		    Thread.sleep( sleepMillis );
		}
                else Thread.yield();
            }
            catch (Exception e)
            {
                reportException( e );
            }
        }
        
	// clean up resources
	conPrint( "Main loop terminating..." );
	try
	{
	    for (int i = 0; i < sound.length; ++i)
	    {
		if (sound[i] != null)
		{
		    sound[i].close();
		    sound[i] = null;
		}
	    }
	    for (int i = 0; i < image.length; ++i)
		image[i] = null;
	}
	catch (Exception e)
	{
	    conPrint( "Exception in stop (" + e.toString() + ")" );
            throw new Error( "Failed to stop!" );
	}

	conPrint( "Closing logfile..." );
	closeLogfile();

	// Notify completion
        done = true;
	
	// is there a better way to kill the parent midlet thread?
	HyperoidMidlet.die();
    }

    /**
     * read all the user preferences strings (optionally resetting to defaults)
     * 
     * @param reset return all settings (and highscore) to defaults.
     */

    private void loadPreferences( boolean reset )
    {
	if (prefs == null)
	    prefs = new AppPreferences( "hyperoid" );

	if (reset)
	    prefs.reset();

	enable_snd = prefs.get( PREFKEY_SOUND, String.valueOf( true ) ).equalsIgnoreCase( "true" );

	String lhsn = prefs.get( PREFKEY_LASTHSNAME, String.valueOf( lastHighScoreName ) );
	lhsn.getChars( 0, 3, lastHighScoreName, 0 );
	int default_score = 1000;
	for (int i = gameHighScore.length - 1; i >= 0; --i)
	{
	    String hs = prefs.get( PREFKEY_HS + i, "EAH," + default_score );
	    conPrint( PREFKEY_HS + i + "=" + hs );
	    int j = hs.indexOf( ',' );
	    if ((j > 0) && (j < (hs.length() - 2)))
	    {
		gameHighScore[i] = new HighScore( hs.substring( 0, j ), Integer.parseInt( hs.substring( j + 1 ) ) );
	    }
	    else
	    {
		gameHighScore[i] = new HighScore( "EAH", default_score );
	    }
	    AttractMenuHighScoreDesc[FIRST_HIGHSCORE_MENU_ENTRY + i] = gameHighScore[i].getMenuText();
	    default_score += 1000;
	}
    }

    /**
     * show what caused the exception.
     * 
     * @param e an Exception.
     */

    private final void reportException( Exception e )
    {
        conPrint( e.getMessage() );
        conPrint( e.toString() );
        e.printStackTrace();
    }

    /**
     * stop the game thread.
     */

    public void stop()          { running = false; }

    /**
     * pause the game thread (not yet implemented).
     */

    public void pause()         {}

    /**
     * query thread run status.
     * @return true if thread is still running.
     */

    public boolean isRunning()  { return( running ); }

    /**
     * query thread done status
     * @return true if thread has not yet cleaned up and exited.
     */

    public boolean isDone()     { return( done ); }
}
